from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint #, shuffle, sample
from functools import reduce
from operator import __or__
from multiprocessing import Pool
import time

import numpy as np
import pytest

from mmgroup import Xsp2_Co1, PLoop, AutPL, Cocode, MM0, MM, XLeech2
from mmgroup.generators import gen_leech2_op_word
from mmgroup.clifford12 import xsp2co1_leech2_count_type2


#####################################################################
# Create test matrices
#####################################################################


def rand_xsp2co1_elem():
    return MM0([(x,'r') for x in "dxpylpylpylpy"])

def rand_xsp2_vectors(length):
    return [randint(0, 0x1ffffff) for i in range(length)]

def create_conjugate_data():  
    # Testing with following test data verfies that generators
    # :math:`\xi^e, e = 1,2` operate on the group :math:`Q_{x0}`
    # correctly. This verifies the correct implementation of 
    # these generator up to sign. The sign can be verified by
    # checking that the generator has order 3
    for exp in (1, 2): 
        g_mm = MM0('l', exp)
        xs = [1 << i for i in range(24)]
        yield g_mm, xs         
    # test with some more random data
    for i in range(20):
        g = rand_xsp2co1_elem()
        xs = rand_xsp2_vectors(8)
        yield g, xs



#####################################################################
# Test conjugtion of extraspecial elements in group G_{x1}
#####################################################################

Co_1_ORDERS = set(list(range(1,17)) + 
      [18,20,21,22,23,24,26,28,30,33,35,36,39,40,42,60])
Gx0_ORDERS = set([x*y for x in Co_1_ORDERS for y in [1, 2, 4]])


def conj_x_by_word(x, g_mm):
    """Conugate x ith a word of elments in the Clifford group

    Here :math:`x` is an element of the normal subgroup :math:`Q_x0`
    of :math:`G_x0`, and :math:`g_{mm}` is an element of the 
    subgroup :math:`G_x0` of the monster.

    :math:`x` is given as an integer in **Leech lattice
    represention** and :math:`g_{mm}` is given as an element
    of the group  object ``mmgroup.MM``, which is the standard 
    instance of the monster group.

    The function returns :math:`g_{mm}^{-1} x g_{mm}` as an
    integer in **Leech lattice represention**.
    """
    w_g = g_mm.mmdata
    return gen_leech2_op_word(x, w_g, len(w_g))

    
@pytest.mark.xsp2co1
def test_xsp2_conjugate(verbose = 0):
    """Test the conjugation of Pauli matrix with unitary matrix"""
    l0, l1, l2 = Xsp2_Co1(), Xsp2_Co1('l', 1),  Xsp2_Co1('l', 2)
    assert l1**2 == l2
    assert l1**3 == l0    
    for ntest, (g_mm, xs) in enumerate(create_conjugate_data()):
        g = Xsp2_Co1(g_mm)
        xs_g_all = [Xsp2_Co1.group.from_xsp(x) for x in xs]
        xs_all = [x.as_xsp() for x in xs_g_all]
        ok = xs == xs_all
        o = g.order()
        assert o in Gx0_ORDERS
        if verbose or not ok:
            print("\nTest %d:\ng = %s" % (ntest, g))
            print("g has order %d" % o)
            print("v =", [hex(x) for x in xs])
            if not ok:
                print("Input and recomputed xsp2 vectors v:")
                print("v =", [hex(x) for x in xs_all])
                err = "Error in recomputation of extraspecial vector"
                raise ValueError(err)

        conj = g.xsp_conjugate(xs)
        ref_conj = [(g**-1 * x * g).as_xsp()  for x in xs_g_all]
        ref_conj1 = [conj_x_by_word(x, g_mm) for x in xs]
        ref_ok = ref_conj == ref_conj1
        if not ref_ok:
            print("Conjugation of v with g =\n", g)
            print("v =", [hex(x) for x in xs])
            print("Reference for v * g via direct calculation")
            print([hex(x) for x in ref_conj]) 
            print("Reference using C function gen_leech2_op_word")
            print([hex(x) for x in ref_conj1]) 
            err = "Error in reference for conjugation in group G_{x1}"
            raise ValueError(err)

        ok = ref_ok and conj == ref_conj1

        if verbose or not ok:
            if not verbose:
                print("Conjugation of v with g =\n", g)
            print("v =", [hex(x) for x in xs])
            print("Conjugated expected:\n", [hex(x) for x in ref_conj])
            if not ok:
                print("Conjugated obtained:\n", [hex(x) for x in conj])
                err = "Error in conjugation in group G_{x1}"
                raise ValueError(err)

        conj_nosign = g.xsp_conjugate(xs, sign = False)
        mask = (1 << 24) - 1
        assert conj_nosign == [x & mask for x in conj]










r"""We deal with the representation of elements of the monster group. 

Generators of the monster group :math:`\mathbb{M}`
..................................................

Conway :cite:`Con85` has constructed a ``196884``-dimensional rational 
representation :math:`\rho` of the monster :math:`\mathbb{M}` based on 
representations of two subgroups 
:math:`G_{x0} = 2_+^{1+24}.\mbox{Co}_1` and 
:math:`N_0 = 2^{2+11+22}.( M_{24} \times S_3)` of 
:math:`\mathbb{M}` which generate :math:`\mathbb{M}`. 
Here :math:`G_{x0}` has a normal extraspecial ``2``-subgroup 
:math:`2_+^{1+24}` with factor group :math:`\mbox{Co}_1`, where
:math:`\mbox{Co}_1` is the automorphism group of the Leech lattice
modulo ``2``. The group :math:`N_0` has a normal subgroup
:math:`2^{2+11+22}`, which is a certain ``2`` group and the factor
group is a direct product of the Mathieu group :math:`M_{24}`
and the symmetric permutation group :math:`S_3` of ``3`` elements.

The group :math:`N_{x0} = N_0 \cap G_{x0}` has index ``3`` in
:math:`N_{0}` and structure :math:`2^{1+24} . 2^{11} . M_{24}`.
It is generated by elements  :math:`x_\delta`, :math:`x_\pi`, 
:math:`x_e`, :math:`y_e`, :math:`z_e`, for all
:math:`x_\delta \in \mathcal{C}^*`, 
:math:`\pi \in {{\rm Aut}_{{\rm St}} \mathcal{P}}` and 
:math:`e \in  \mathcal{P}`.

Here  :math:`\mathcal{C}^*` is the Golay cocode defined in
section  :ref:`golay-label`, :math:`\mathcal{P}` is the Parker
loop defined in section  :ref:`parker-loop-label`, and 
:math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` is the automorphism 
group of  :math:`\mathcal{P}` defined in section  
:ref:`aut_ploop_label`.
The group :math:`N_{0}` has a subgroup isomorphic to
:math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` generated by the
generators  :math:`x_\delta, \delta \in \mathcal{C}^*`
and :math:`x_\pi,  \pi \in {{\rm Aut}_{{\rm St}} \mathcal{P}}`.
The generators  :math:`x_\delta` generate the subgroup of diagonal
automorphisms of :math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}`. 


The group :math:`N_{0}` is generated by :math:`N_{x0}` and by 
another element :math:`\tau` of :math:`N_{0}` or order ``3``. The
group :math:`G_{x0}` is generated by :math:`N_{x0}` and by another
element  :math:`\xi` of :math:`G_{x0}` or order ``3``. The elements
:math:`x_\delta`, :math:`x_\pi`, :math:`x_e`, 
:math:`y_e`, :math:`z_e`,  :math:`\tau` and  :math:`\xi`  
generate the monster group  :math:`\mathbb{M}`.
In this  package we use the definitions of the generators 
in  :cite:`Seysen20`, which incorporate a modification of 
:cite:`Con85` made in :cite:`Iva09`. 
This leads to simpler relations in  :math:`N_{0}`.
The generators :math:`x_e`, :math:`y_e`, and :math:`z_e` in  
:cite:`Seysen20` correspond to the generators
:math:`x_e \cdot z_{-1}^{|e/4|}`, :math:`y_e \cdot x_{-1}^{|e/4|}`,
and  :math:`z_e \cdot y_{-1}^{|e/4|}`` in :cite:`Con85`.
 

Creating an instance of the monster group
.........................................


For dealing with the monster group :math:`\mathbb{M}` the user must 
first create an instance ``MM`` of class |MMGroup|, which 
represents the monster group :math:`\mathbb{M}`::

      MM = MMGroup()


Class |MMGroup| is implemented as a singleton. Calling the constructor
of that class always returns the same instance of that class. There is
the predefined instance ``mmgroup.MM`` of class |MMGroup|.

Elements of the monster group are created by calling that instance ``MM`` 
as a function with a variable number of arguments. Here each argument 
describes an element of :math:`\mathbb{M}` which is (usually) one of the 
generators listed above. The the function returns the product of all 
these elements as an instance of class |MMGroupWord|. Instances of class  
|MMGroupWord| model elements of the Monster group :math:`\mathbb{M}`.

An argument passed to an instance ``MM`` of class  |MMGroup| may be a pair 
``(tag, value)``, where ``tag`` is a single small letter describing the 
type of the generator of :math:`\mathbb{M}`, and ``value`` is an 
integer describing the value of that generator. Alternatively, ``value``
may be an instance of the appropriate algebraic structure used for
indexing a generator of a given type as indicated in the table below.


Implementation of the generators of the monster group
.....................................................


Math papers may use (at least) Latin or Greek letters for labelling
objects, but most programming languages are restricted to ASCII characters.
Assuming that ``MM`` is an instance of class  |MMGroup| representing
a monster group, the following table shows how to create generating
elements of ``MM``: 


.. table:: Construction of generating elements of the monster
  :widths: 10 90


  ===================  ==========================================================
  Element              Construction as an element of ``MM``, 
                       with ``MM`` of type |MMGroup|
  ===================  ==========================================================
  :math:`x_\delta`     ``MM(('d', delta))``, ``delta`` an instance of class
                       |Cocode|; 

                       ``MM(('d', delta))`` returns 
                       ``MM('d', Cocode(delta))`` for ``0 <= delta < 0x1000``. 

  :math:`x_\pi`        ``MM(pi)``, ``pi`` an instance of class |AutPL|;

                       ``MM(('d', delta), ('p', n))`` returns 
                       ``MM(AutPL(delta, n))`` 
                       for ``0 <= delta < 0x1000``, ``0 <= n < 244823040`` .

  :math:`x_e`          ``MM(('x', e))``, ``x`` an  instance of class |PLoop|;

                       ``MM(('x', e))`` returns ``MM(('x', PLoop(e)))`` for
                       ``0 <= e < 0x2000``.

  :math:`y_e`          ``MM(('y', e))``,  ``e`` as in case :math:`x_e`. 

  :math:`z_e`          ``MM(('z', e))``,  ``e`` as in case :math:`x_e`. 

  :math:`\tau^e`       ``MM(('t', e))``, exponent  ``e`` is an integer 
                       which is taken modulo ``3``.

  :math:`\xi^e`        ``MM(('l', e))``,  exponent  ``e`` is an integer 
                       which is taken modulo ``3``.
  ===================  ==========================================================


More possibilities for constructing elements of an instance of class 
|MMGroup| are given in the description of that class. An element ``g``
of an instance of the monster group is modelled as in instance of class 
|MMGroupWord|. ``g.group`` is the group where ``g`` belongs to; that 
group is an instance of class |MMGroup|. Multiplication and 
exponentiation of group elements works a usual.  


Multiplication of elements of the monster group
...............................................

Internally, an element of  :math:`\mathbb{M}` is represented as a word
in the generators given above. The user should be aware of the fact
that multiplication with the ``*`` operator is a concatenation of such
words, followed by (rather incomplete) reduction step. This means that 
multplying words may still lead to an exponential growth of the word 
length. 

On can apply the method ``simplify`` to an element of the monster group.
This method shortens a word representing an element of :math:`\mathbb{M}`
to a reasonable length with a very high probability. However, applying 
this method usually takes a long time. On the author's computer it may 
take several minutes to shorten the product of two (previously shortened) 
random words. This is yet considerably faster than the run time reported 
in :cite:`Wilson13`.

In such a word we always reduce substrings of generators of the 
subgroup :math:`N_0` of :math:`\mathbb{M}` to a standard form, 
which is easy. We apply no relations to the remaining generator
:math:`\xi`, except for :math:`\xi^3=1`. Method ``simplify`` uses
the technique described in :cite:`Wil03` for shortening a word
in the monster group.




Future development
..................

Future versions of this package may implement improved reduction
strategies for words of generators of :math:`\mathbb{M}` :


 * Long words of generators of :math:`\mathbb{M}` can be shortened 
   with high probability by method ``mmgroup.MMGroupWord.simplify``. 
   This method uses the algorithm in :cite:`Wil03`. This may take 
   a long time. Here improvements in future versions are desirable.

"""
# References in the __docstr__ see file docs/source/references.bib


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



import sys
import os
import collections
import re
import warnings
from numbers import Integral
import numpy as np
from random import randint, sample


from mmgroup.generators import rand_get_seed
from mmgroup.generators import gen_rng_modp
from mmgroup.structures.parse_atoms import ihex, TaggedAtom
from mmgroup.structures.abstract_group import singleton
from mmgroup.structures.abstract_mm_group import AbstractMMGroupWord
from mmgroup.structures.abstract_mm_group import AbstractMMGroup
from mmgroup.structures.parse_atoms import  AtomDict
from mmgroup.structures.construct_mm import iter_mm       
from mmgroup.structures.construct_mm import load_group_name     
from mmgroup.generators import gen_leech2_reduce_type4
from mmgroup.generators import mm_group_invert_word
from mmgroup.clifford12 import xsp2co1_check_word_g_x0 
from mmgroup.clifford12 import xsp2co1_reduce_word      
from mmgroup.clifford12 import xsp2co1_traces_fast      
from mmgroup.clifford12 import chk_qstate12
from mmgroup.clifford12 import xsp2co1_rand_word_G_x0
from mmgroup.clifford12 import xsp2co1_rand_word_N_0
from mmgroup.mm import mm_vector



try:
    from mmgroup.mat24 import MAT24_ORDER, pow_ploop, ploop_theta
except (ImportError, ModuleNotFoundError):
    from mmgroup.dev.mat24.mat24_ref import Mat24    
    MAT24_ORDER = Mat24.MAT24_ORDER
    pow_ploop = Mat24.pow_ploop
    ploop_theta = Mat24.ploop_theta
    del Mat24


from mmgroup.structures.ploop import Cocode, PLoop
from mmgroup.structures.autpl import StdAutPlGroup, autpl_from_obj, AutPL

from mmgroup.generators import mm_group_mul_words






###########################################################################
# Import functions from module ``mmgroup.mm_order`` on demand
###########################################################################


# Functions to be imported from module mmgroup.mm_order
check_mm_order = None
check_mm_equal = None
check_mm_half_order = None
check_mm_in_g_x0 = None

def import_mm_order_functions():
    """Import functions from module ``mmgroup.mm_order``.

    We import these functions from module ``mmgroup.mm_order``
    on demand. This avoids an infinite recursion of imports.
    """
    global check_mm_order, check_mm_equal
    global check_mm_half_order, check_mm_in_g_x0
    global mm_reduce_M
    from mmgroup.structures.mm_order import check_mm_order
    from mmgroup.structures.mm_order import check_mm_equal 
    from mmgroup.structures.mm_order import check_mm_half_order
    from mmgroup.structures.mm_order import check_mm_in_g_x0




###########################################################################
# Import functions from module ``mmgroup.structures.xsp2_co1`` on demand
###########################################################################


# Functions to be imported from modules mmgroup.structures.xsp2_co1
# and mmgroup.structures.involutions
Xsp2_Co1 = None
xsp2co1_to_mm = None
mm_conjugate_involution = None
reduce_via_power = None


def import_Xsp2_Co1():
    global Xsp2_Co1, xsp2co1_to_mm, mm_conjugate_involution
    global reduce_via_power
    from mmgroup.structures.xsp2_co1 import Xsp2_Co1 as f
    Xsp2_Co1 = f
    from mmgroup.structures.xsp2_co1 import Xsp2_Co1 as f
    xsp2co1_to_mm = f
    from mmgroup.structures.involutions import mm_conjugate_involution as f
    mm_conjugate_involution = f    
    from mmgroup.structures.involutions import reduce_via_power as f
    reduce_via_power = f


###########################################################################
# Word class for the group MM0
###########################################################################





class MM0(AbstractMMGroupWord):
    r"""Models an element of the monster group :math:`\mathbb{M}`

    Let ``MM`` be an instance of class ``MMGroup``, and let ``g1``, 
    ``g2`` be elements of ``MM``.  Then
    ``g1 * g2``  means group multiplication, and ``g1 ** n`` means
    exponentiation of ``g1`` with the integer ``n``. ``g1 ** (-1)`` 
    is the inverse of ``g``. ``g1 / g2`` means ``g1 * g2 ** (-1)``.
    We have ``1 * g1 == g1 * 1 == g1`` and ``1 / g1 == g1 ** (-1)``.

    ``g1 ** g2`` means ``g2**(-1) * g1 * g2``.   

    Let ``V`` be a vector space that is a representation of ``MM``,
    see class |MMSpace| for details. An element ``g1`` of ``MM`` 
    operates on the vector space  ``V`` by right multiplication.  

    :var group:
        This attribute contains the group to which the element belongs.
        That group is an instance of class |MMGroup|.

    .. warning::
       The constructor of this class is not for public use! You
       may call an instance ``MM`` of class  |MMGroup| for
       constructing elements of the instance ``MM`` of the monster 
       group.
  
    """
    ERR_ITER = "A monster group element g is not iterable. Use g.mmdata instead"
    MIN_LEN = 16
    __slots__ =  "length", "_data", "reduced"
    def __init__(self,  tag = None, atom = None, *args, **kwds):
        self.reduced = 0
        atoms = iter_mm(self.group, tag, atom)
        self._data = np.fromiter(atoms, dtype = np.uint32) 
        self.length = len(self._data)
        self._extend(self.MIN_LEN)
        
                  
    def _extend(self, length):
        len_ = len(self._data)
        if length > len_:
            ap = np.zeros(max(length - len_, len_), dtype = np.uint32)
            self._data = np.append(self._data, ap)
             
    @property
    def mmdata(self):
        """Return the internal representation of the group element

        Internally, an element of the monster group is represented
        as an array of unsigned 32-bit integers, where each entry
        of the array describes a generator. See section
        :ref:`header-mmgroup-generators-label` for details.
        """
        return np.copy(self._data[:self.length])


    def __bool__(self):
        return True
        
    def __len__(self):
        raise TypeError(self.ERR_ITER)

    def _setdata(self, data):
        len_ = len(data)
        self._extend(len_)
        self._data[:len_] = data
        self.length = len_
        self.reduced = False

        
    def __getitem__(self,i):
        raise TypeError(self.ERR_ITER)
 
        
    def is_reduced(self):
        """Return ``True`` if the element of the monster group is reduced

        An element ``g`` of the monster group represented by an instance
        of this class may be reduced by calling
        ``g.reduce()``.
        """
        return self.length == self.reduced




    def order(self, max_order = 119):
        r"""Return the order of the element of the monster group

        We use the method in :cite:`LPWW98`, section 7, for computing
        the order of an element of the monster.

        If the argument ``max_order`` is present then the order of the 
        element is checked up to (and including) ``max_order`` only.  
        Then the function returns ``0`` if the order is greater than 
        ``max_order``. By default, the function returns the exact 
        order of the element.
        """
        if check_mm_order is None:
            import_mm_order_functions()
        return check_mm_order(self, max_order)

    def half_order(self, max_order = 119):
        r"""Return the (halved) order of the element of the monster group

        The function  returns a pair ``(o, h)``, where ``o`` is the 
        order of the element.

        If ``o`` is even then the function returns the pair ``(o, h)``,
        where ``h`` is the element raised to to the ``(o/2)``-th power. 
        Otherwise the function returns the pair ``(o, None)``.
   
        Parameter ``max_order`` is as in function ``check_mm_order``. 

        If ``h`` is in the subgroup :math:`G_{x0}`` then ``h`` is 
        returned as a word in the generators of that subgroup.
        """
        if check_mm_half_order is None:
            import_mm_order_functions()
        return check_mm_half_order(self, max_order)

    def in_G_x0(self):
        r"""Check if the element is in the subgroup :math:`G_{x0}`

        The function returns True if this is the case. If this is
        the case then the element is converted to a word in the
        generators of :math:`G_{x0}`.

        This method uses geometric information about the Leech 
        lattice taken from :cite:`Iva99`.
        """
        if check_mm_in_g_x0 is None:
            import_mm_order_functions()
        return check_mm_in_g_x0(self) is not None
                          
    def chi_G_x0(self, involution = None):
        r"""Compute characters of element of subgroup :math:`G_{x0}`

        If the element is in the subgroup :math:`G_{x0}` then the 
        function returns a tuple 
        :math:`(\chi_M, \chi_{299}, \chi_{24}, \chi_{4096})`
        of integers. Otherwise it raises ``ValueError``.

        Here :math:`\chi_M` is the character of the element in the
        196833-dimensional rep :math:`198883_x` of the monster.

        By Conway's construction of the monster we have:

        :math:`198883_x =  299_x \oplus 24_x \otimes  4096_x
        \oplus 98280_x`,

        for suitable irreducible representations 
        :math:`299_x, 24_x, 4096_x, 98280_x` of the group 
        :math:`G_{x0}`. The corresponding characters of the
        element of  :math:`G_{x0}` are returned in the tuple
        given above.  

        While the product :math:`\chi_{24} \cdot \chi_{4096}`
        is well defined, the factors  :math:`\chi_{24}` and 
        :math:`\chi_{4096}` are defined up to sign only. We
        normalize these factors such that the first nonzero value 
        of the pair :math:`(\chi_{24}, \chi_{4096})` is positive. 

        If parameter ``involution`` is a 2B involution :math:`i` then
        the element :math:`g` given by ``self`` must centralize that
        involution  :math:`i`. In this case we compute the  
        corresponding  characters of  :math:`g^h` for a  
        :math:`h \in \mathbb{M}` such that  :math:`i^h`  is the
        central involution of :math:`G_{x0}`.
        """
        if involution:
             _, h = involution.conjugate_involution()
             g = self**h
             if g.in_G_x0() is None:
                 err = "Element does not centralize the given 2B involution"
                 raise ValueError(err)
        else:
             g = self
             if g.in_G_x0() is None:
                 err = "Element is not in the subgroup G_x0 of the monster"
                 raise ValueError(err)

        if Xsp2_Co1 is None: 
            import_Xsp2_Co1()
        elem = Xsp2_Co1(g)

        a = np.zeros(4, dtype = np.int32)
        res = chk_qstate12(xsp2co1_traces_fast(elem._data, a))
        chi24, chisq24, chi4096, chi98260 = map(int, a[:4])
        chi299 = (chi24**2 + chisq24) // 2 - 1
        chi_M = chi299 + chi98260 + chi24 * chi4096
        return chi_M, chi299, chi24, chi4096
       
    def in_N_x0(self):
        r"""Check if the element is in the subgroup :math:`N_{x0}`

        The function returns True if this is the case. If this is
        the case then the element is converted to a word in the
        generators of :math:`N_{x0}`.
        """
        if check_mm_in_g_x0 is None:
            import_mm_order_functions()
        if check_mm_in_g_x0(self) is None:
            return False
        self.reduce()
        for atom in self.mmdata:
            if ((atom >> 28) & 7) > 4:
                return False
        return True
 
    def in_Q_x0(self):
        r"""Check if the element is in the subgroup :math:`Q_{x0}`

        The function returns True if this is the case. If this is
        the case then the element is converted to a word in the
        generators of :math:`Q_{x0}`.
        """
        if check_mm_in_g_x0 is None:
            import_mm_order_functions()
        if check_mm_in_g_x0(self) is None:
            return False
        self.reduce()
        for atom in self.mmdata:
            if (atom & 0x70000000) not in [0x10000000,  0x30000000]:
                return False
        return True
             

    def conjugate_involution(self, check=True, ntrials=40, verbose=0):
        r"""Find an element conjugating an involution into the centre

        If the element :math:`g` given by ``self`` is an involution 
        in  the monster then the method computes an element :math:`h` 
        of the monster   with  :math:`h^{-1} g h = z`, where  
        :math:`z` is defined as follows:

        If :math:`g = 1`, we put :math:`h = z = 1`

        if :math:`g` is a 2A involution (in the monster) then we let
        :math:`z` be the involution in  :math:`Q_{x0}` corresponding 
        to the Golay cocode word with entries  :math:`2,3` being set.

        if :math:`g` is a 2B involution (in the monster) then we let 
        :math:`z` be the central involution in :math:`G_{x0}`

        The function returns the element :math:`h` of the monster. It
        raises ``ValueError`` if :math:`g` is not an involution in
        the monster, or no suitable element  :math:`h` can be foound. 

        This function may take a long time. Parameter  ``ntrials``
        gives the number of trials to find a suitable element
        :math:`h`. Default is  ``ntrials = 40``. The function may
        fail after that number of trials even if the element is a
        2B involution.

        If parameter ``check`` is True (default) then the function 
        first checks if ``g`` is an involution.

        In future versions support for multiprocessing may be added.
        """
        if mm_conjugate_involution is None: 
            import_Xsp2_Co1()
        return mm_conjugate_involution(self, check, ntrials, verbose)

    def half_order_chi(self,  ntrials=40):
        r"""Return order and some character information of the element

        The function  returns a triple ``(o, chi, h)``, where ``o`` is 
        the order of the element :math:`g` of the monster given 
        by ``self``. 

        If the order :math:`o` of  :math:`g` is odd then the triple 
        ``(o, None, None)`` is returned.

        If  the order of :math:`g` is even then we return the triple
        ``(o, chi, h)``. Here ``h`` is an element :math:`h` of
        :math:`\mathbb{M}` such that :math:`z = h^{-1} g^{o/2} h` is 
        the  standard 2A or 2B involution as in method
        ``conjugate_involution``.

        If :math:`g` powers up to a 2B involution then ``chi``
        is equal to the character of  :math:`u = h^{-1} g h`. 
        Here the character of the element :math:`u` of
        :math:`G_{x0}` is returned  as a quadruple as in method
        ``chi_G_x0``. 

        If :math:`g` powers up to a 2A involution then ``chi``
        is equal to to ``None``.
        """
        o, ho = self.half_order()
        chi, h = None, None
        if ho:
            _, h = ho.conjugate_involution(True, ntrials)
            try:
               chi = (self**h).chi_G_x0()
            except ValueError:
               chi = None
        return o, chi, h



    def simplify(self, ntrials=None, verbose=0):
        """Try to simplify an element of the monster group

        This function tries to simplify an element of the monster group.
        It may take a long time, but this is the only way to prevent the
        explosion of the word lengths of elements of the monster group. 

        The current version uses the word shortening algorithm
        in  :cite:`Wil03`. Parameter ``ntrials`` specifies the number
        of trials. Here the default value for ``ntrials`` should be
        used, since the parameter may be dropped in future versions.

        The word shortening algorithm and the implementation of this 
        method is experimental and may change in future versions!
        """
        if check_mm_in_g_x0 is None:
            import_mm_order_functions()
        check_mm_in_g_x0(self)
        self.reduce()
        weight = sum([((x >> 28) & 7) == 5 for x in self.mmdata])
        #print(weight)
        if weight <= 9:
             return self
        if ntrials is None:
            ntrials = 40
        else:
            w = ("The default value of parameter 'ntrials' should be used"
                 " in method mmgroup.MM.simplify")
            warnings.warn(w)
        if reduce_via_power is None:
            import_Xsp2_Co1()
        try: 
            g = reduce_via_power(self, ntrials, verbose = verbose)
            assert g == self
            self._setdata(g.mmdata)
            self.reduce()
        except:
            w = "Simplification of monster group element failed"
            warnings.warn(w)
        return self


###########################################################################
# The class representing the group MM
###########################################################################




@singleton
class MM0Group(AbstractMMGroup):
    r"""An instance ``MM`` of this class models an instance of the monster

    :param: None

    :return: An instance of the monster group 
    :rtype:  an instance of class |MMGroup|

    This means that ``MM = MMGroup()`` creates an instance ``MM`` of the 
    monster group  :math:`\mathbb{M}`. For generating an element ``g`` 
    of ``MM`` one must call ``MM`` as a function with a variable number 
    of arguments. Depending on its type, each argument is evaluated to
    an element of ``MM`` as indicated in the table below, and the
    product of these elements is returned. 

    Elements of the monster group are implemented as instances of class
    ``MMGroupWord``. The preferred way to construct an element of the 
    monster group is to call an instance of class |MMGroup|.  

    .. table:: Legal types for constructing an element of the monster
      :widths: 25 75

      +---------------------+-------------------------------------------+
      | Object of type      | Evaluates to                              |
      +=====================+===========================================+
      | tuple (``tag, i``)  | ``MM((tag, i))`` is equivalent to         |
      |                     | ``MM.atom(tag, i)``                       |
      +---------------------+-------------------------------------------+
      | class |PLoop|       | The Parker loop element :math:`d`         | 
      |                     | given by that object is mapped to         |
      |                     | :math:`x_d \in \mathbb{M}`                |
      +---------------------+-------------------------------------------+
      | class |AutPL|       | The Parker loop automorphism :math:`\pi`  | 
      |                     | given by that object is mapped to         |
      |                     | :math:`x_\pi \in \mathbb{M}`              |
      +---------------------+-------------------------------------------+
      | class |Cocode|      | The Golay cocode element :math:`\delta`   | 
      |                     | given by that object is mapped to         |
      |                     | :math:`x_\delta \in \mathbb{M}`           |
      +---------------------+-------------------------------------------+
      | class |MMGroupWord| | A deep copy of the given element of       | 
      |                     | :math:`\mathbb{M}` is made                |
      +---------------------+-------------------------------------------+
      | ``str``             | For an element ``g`` of ``MM`` we have    |
      |                     | ``MM(str(g)) == g``.                      |
      |                     |                                           |
      |                     | So we can reread                          |
      |                     | printed elements of ``MM``.               |
      +---------------------+-------------------------------------------+

    See class |MMGroupWord| for the group operation on an instance
    ``MM`` of  class |MMGroup|.

    Two elements ``g1, g2`` of the monster group can be tested for 
    equality with the ``==`` operator as usual. Here we use the 
    method given in :cite:`Wil03` for checking ``g1 * g2**(-1) == 1``.

    Elements ``g1``, ``g2`` that belong to different instances of
    class |MMGroup| are considered unequal.
    """
    word_type = MM0
    #STR_FORMAT = r"M<%s>"
    group_name = "M0"


    def __init__(self):
        """ TODO: Yet to be documented     


        """
        super(MM0Group, self).__init__()

    def reduce(self, g1, copy = False):
        l1 = g1.length
        if g1.reduced < l1:
            if copy:
                g1 = self.copy_word(g1)
            l_tail = l1 - g1.reduced
            g1._extend(l1 + l_tail + 1)
            g1._data[l1 : l1 + l_tail] = g1._data[g1.reduced : l1]
            tail =  g1._data[l1:]
            l1 = mm_group_mul_words(g1._data, g1.reduced, tail, l_tail, 1)
            g1.reduced = g1.length = l1
        return g1

        
    def _imul(self, g1, g2):
        l1, l2 = g1.length, g2.length
        g1._extend(2*(l1 + l2) + 1)
        g1._data[l1 : l1 + l2] = g2._data[:l2]
        l1 += l2
        l_tail = l1 - g1.reduced
        g1._data[l1 : l1 + l_tail] = g1._data[g1.reduced : l1]
        tail = g1._data[l1:]
        l1 = mm_group_mul_words(g1._data, g1.reduced, tail, l_tail, 1)
        g1.reduced = g1.length = l1
        return g1

    def _invert(self, g1):
        w = self.word_type()
        w._setdata(np.flip(g1.mmdata) ^ 0x80000000)
        return self.reduce(w)

    def copy_word(self, g1):
        result = self.word_type()
        result._setdata(g1.mmdata)
        result.reduced = g1.reduced
        return result

    def _equal_words(self, g1, g2):
        if not check_mm_equal is None:
            return g1.group == g2.group and check_mm_equal(g1, g2)
        try:
            import_mm_order_functions()
            return g1.group == g2.group and check_mm_equal(g1, g2)
        except:
            if g1.group != g2.group:
                return False
            g1.reduce()
            g2.reduce()
            if (g1.mmdata == g2.mmdata).all():
                 return True
            raise ValueError("Don't know if monster group elements are equal")



StdMM0Group = MM0Group()
MM0.group = StdMM0Group
load_group_name(StdMM0Group, "M0")







from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint #, shuffle, sample
from functools import reduce
from operator import __or__
from multiprocessing import Pool
import time

import numpy as np
import pytest

from mmgroup import Xsp2_Co1, PLoop, AutPL, Cocode, MM0, MM, XLeech2
from mmgroup.generators import gen_leech2_op_word
from mmgroup.clifford12 import xsp2co1_set_elem_word_scan
from mmgroup.clifford12 import xsp2co1_unit_elem
from mmgroup.structures.construct_mm import iter_mm


#####################################################################
# Create test matrices
#####################################################################



def construct_xsp2co1_testcases():
    def a(string_data):
        data = [(x, 'n') for x in string_data]
        return np.array(list(iter_mm(None, data)), dtype = np.uint32)

    for tag0 in "pdxyl":
        yield a(tag0), 1
        for tag1 in "pdxyl":
            yield a(tag0 + tag1 + 't'), 2

    for i in range(5):
        data = Xsp2_Co1('r', 'G_x0').mmdata
        yield data, len(data)

def dirty_elem():
    return np.array([randint(0, 0xffffffffffffffff) for i in range(26)],
        dtype = np.uint64)

@pytest.mark.xsp2co1
def test_xsp2co1_set_elem_word_scan():
    for a, len_a in construct_xsp2co1_testcases():
        elem0 = dirty_elem()
        elem1 = dirty_elem()
        s0 =  xsp2co1_set_elem_word_scan(elem0, a, len(a), 0)
        assert s0 == len_a
        xsp2co1_unit_elem(elem1)
        s1 =  xsp2co1_set_elem_word_scan(elem1, a, len(a), 1)
        assert s1 == len_a
        assert s0 == s1
        xsp = Xsp2_Co1()
        np.copyto(xsp._data, elem0)
        assert Xsp2_Co1('a', a[:len_a]) == xsp









from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


"""Modelling an abstract group

Yet to be documented

"""
import sys
import re
import warnings
from copy import deepcopy
from random import sample, randint
from numbers import Integral, Number


from mmgroup.structures.abstract_group import AbstractGroup
from mmgroup.structures.abstract_group import AbstractGroupWord
from mmgroup.structures.construct_mm import iter_mm
from mmgroup.structures.construct_mm import iter_strings_from_atoms
from mmgroup.structures.construct_mm import iter_tuples_from_atoms



####################################################################
####################################################################
### Class AbstractGroup and helpers for that class
####################################################################
####################################################################

####################################################################
### Class AbstractGroupWord
####################################################################


class AbstractMMGroupWord(AbstractGroupWord):
    """Model an element of an abstract group.

    This is an abstract base class. Subclasses of this class
    correspond to (preimages of) susbgroups of the monster generated
    by some or all of the standard generators of the monster group.
    Instances of such subclasses are elements of the corresponding 
    groups.

    The standard group operations '*' '/' (=right multiplication with
    the inverse) and '**' are implemented here.

    g1 ** g2  means  g2**(-1) * g1 * g2  for group elements g1, g2.

    Here a group is an instance of (a subclass of) class AbstractGroup. 

    For each word a group 'g' should be passed as a keyword argument 
    'group = g'. If a class of type 'AbstractGroup' contains one 
    instance only, the corresponding subclass of this class may 
    contain a class attribute 'group' referring to that group. Then 
    the user  may contruct elements of that group using the 
    constructor of that subclass of this class.
    """
    group_name = "AbstractMMGroup"
    def __init__(self, tag = None, atom = None, reduce = True):
        self.from_data(self.group, tag, atom)
        if reduce:
            self.reduce()
 


    def reduce(self, copy = False):
        """Reduce a group element

        If group elements are implemented as words, some functions
        may produce unreduced words. This function  reduces the
        group element in place.

        Note that all operators return reduced words. Functions return
        reduced words unless stated otherwise. However, reducing all
        words representing the same group element to the same word may 
        be beyond the capabilties of a program. 

        If ``copy`` is set then a reduced copy of the element is 
        returned, in case that the input element is not already 
        reduced.
        """
        return self.group.reduce(self, copy)

    @property
    def mmdata(self):
        """Return the internal rpresentation of the element

        This method returns the internal representation of an element
        of the monster group as a numpy array of unsigend 32-bit
        integers.

        The internal representation is described in section 
        *The monster group* in the *API reference*.
        """
        raise NotImplementedError("Abstract class")


    def str(self):
        """Represent group element as a string"""
        try:
            return self.group.str_word(self)
        except NotImplementedError:
            return super(AbstractGroupWord, str)()
    __repr__ = str

    def raw_str(self):
        """Represent group element as a string

        In contrast to the standard method ``str``, this method
        returns a string containing the raw data of the group
        element only.
        """
        return self.group.raw_str_word(self)


    def as_tuples(self):
        """Convert group element to a list of tuples

        For a group element ``g`` the following should hold:

        ``self.__class__(self.as_tuples()) == self`` .

        This shows how to convert a group element to a list of tuples
        and vice versa.
        """
        return list(iter_tuples_from_atoms(self.mmdata))



def is_mmgroup_word(g):
    return isinstance(g,AbstractMMGroupWord) and g.group.is_mmgroup 

####################################################################
### Class AbstractGroup
####################################################################


class AbstractMMGroup(AbstractGroup):
    """Model an abstract group"""
    word_type = AbstractMMGroupWord  # type of an element (=word) in the group
    group_name = "AbstactMMGroup"
    is_mmgroup = True

    def __init__(self, *data, **kwds):
        """Creating instances is only possible for concrete groups

         
        """

    ### The following methods must be overwritten ####################

    def atom(self, *args):
        """Return an atomic element of this group. 

        Calling this function without any arguments should return
        the neutral element of this group.
        """
        raise NotImplementedError("No atoms defined in abtract group")  

    def _imul(self, g1, g2):
        """Return product g1 * g2 of group elements g1 and g2.

        g1 may be destroyed but not g2.

        This method is called for elements g1 and g2 of the group
        'self' only. It should return the reduced product.
        """
        raise NotImplementedError("No multiplication in abstract group")

    def _imul_nonreduced(self, g1, g2):
        """ Non-reduced multiplication g1 * g2

        The result of this mathod the product g1 * g2. If group 
        elements are represented by words then the function returns  
        the concatenation of the words g1 and g2 without reducing it.

        Here g1 and g2 must be elements of the same group.

        The default implementation does not distinguish between
        non-reduced and reduced multipLication. 

        g1 may be destroyed but not g2.

        This method is called for elements g1 and g2 of the group
        'self' only.

        This method is used in method word() of this class for
        constructing a word of the group without reducing it.
        """
        return self._imul(g1, g2) 


    def _invert(self, g1):
        """Return inverse g1**(-1) of group element g1.

        g1 must not be destroyed.

        This method is called for elements g1 of the group
        'self' only. It should return the reduced inverse.
        """
        raise NotImplementedError("No inversion in abstract group")
        
    ### The following methods should be overwritten ###################

    def copy_word(self, g1):
        """Return deep copy of group element ``g1``"""
        g_copy = deepcopy(g1)
        # Even a deep copy of an element is still in the same group
        g_copy.group = g1.group
        return g_copy

    def _equal_words(self, g1, g2):
        """Return True iff elements g1 and g2 are equal 

        This method is called for elements g1 and g2 of the group
        'self' only.
		
        In concrete group this method should be overwritten with
        a comparison of the relevant attributes of g1 and g2.

        Caution:
        Non-reduced words may be considered unequal even if they
        represent the same element. Use g1.reduce() or g1 * 1 to
        obtain the reduced form of g1. See method reduce() for
        details.
        """
        return g1 == g2


    def reduce(self, g, copy = False):
        """Reduce the word ``g`` which is an element of the group

        We assume that the representation of a group element
        is not always given in a unique form that we call
        the reduced form. 

        This method tries to achieve this goal. Group elements
        are reduced by any operator, except for the
        ``==`` and  ``!=`` operators. 

        For test purposes, is is useful to obtain a group 
        element in non-reduced form. Applications should
        create reduced group elements only.

        One way to obtain avoid reduction is to call method 
        ``word()`` of this class with elements separated by 
        commas. Then no reduction takes place across the factors
        separated by commas. 

        If argument ``copy`` is True, a reduced copy of ``g``
        should be returned if ``g`` is not reduced.
        """
        return g


    ### The following methods need not be overwritten #################



    def str_word(self, g):
        """Convert group atom ``g`` to a string

        For an element ``g`` of this group ``g.group.str_word(g)``
        should be equivalent   to ``g.str()``.
        """
        """Represent group element as a string"""
        atoms = g.mmdata
        strings = iter_strings_from_atoms(atoms, abort_if_error=0)
        s = "*".join(strings) 
        if s == "": s = "1"
        return "%s<%s>" % (self.group_name, s)

    def raw_str_word(self, g):
        """Convert group atom ``g`` to a string

        For an element ``g`` of this group ``g.group.raw_str_word(g)``
        should be equivalent   to ``g.raw_str()``.
        """
        """Represent group element as a string"""
        atoms = g.mmdata
        strings = iter_strings_from_atoms(atoms, abort_if_error=0)
        s = "*".join(strings) 
        return "1" if  s == "" else s
                 

    def neutral(self):
        """Return neutral element of the group"""
        assert self.word_type.group == self
        return self.word_type()


    def _embed_number(self, n):
        """Try to embed the number n into the group.

        The function returns the number n as a group element or
        it raises TypeError if this is not possible.
        For a matrix group it may e.g. return n times the unit matrix.

        By default, we only map the number 1 to the neutral element.
        """
        raise ValueError("Nixda!!!") # TODO!!!
        if n == 1:
            return self.neutral()
        raise TypeError("Cannot convert a number to a group element")

   




    ### The following methods should not be overwritten ###############



    def __contains__(self, other):
        """Return True iff 'other' is an element of the group"""
        try:
            if not isinstance(other, self.word_type): 
                 return False
            if other.group != self: 
                 return False
            return True
        except:
            return False


    def __call__(self, *args):
        """Convert args to group element

        """
        assert self.word_type.group == self
        return self.word_type(*args)
    



AbstractMMGroupWord.group = AbstractMMGroup




r"""We deal with the representation of elements of the monster group. 

Generators of the monster group :math:`\mathbb{M}`
..................................................

Conway :cite:`Con85` has constructed a ``196884``-dimensional rational 
representation :math:`\rho` of the monster :math:`\mathbb{M}` based on 
representations of two subgroups 
:math:`G_{x0} = 2_+^{1+24}.\mbox{Co}_1` and 
:math:`N_0 = 2^{2+11+22}.( M_{24} \times S_3)` of 
:math:`\mathbb{M}` which generate :math:`\mathbb{M}`. 
Here :math:`G_{x0}` has a normal extraspecial ``2``-subgroup 
:math:`2_+^{1+24}` with factor group :math:`\mbox{Co}_1`, where
:math:`\mbox{Co}_1` is the automorphism group of the Leech lattice
modulo ``2``. The group :math:`N_0` has a normal subgroup
:math:`2^{2+11+22}`, which is a certain ``2`` group and the factor
group is a direct product of the Mathieu group :math:`M_{24}`
and the symmetric permutation group :math:`S_3` of ``3`` elements.

The group :math:`N_{x0} = N_0 \cap G_{x0}` has index ``3`` in
:math:`N_{0}` and structure :math:`2^{1+24} . 2^{11} . M_{24}`.
It is generated by elements  :math:`x_\delta`, :math:`x_\pi`, 
:math:`x_e`, :math:`y_e`, :math:`z_e`, for all
:math:`x_\delta \in \mathcal{C}^*`, 
:math:`\pi \in {{\rm Aut}_{{\rm St}} \mathcal{P}}` and 
:math:`e \in  \mathcal{P}`.

Here  :math:`\mathcal{C}^*` is the Golay cocode defined in
section  :ref:`golay-label`, :math:`\mathcal{P}` is the Parker
loop defined in section  :ref:`parker-loop-label`, and 
:math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` is the automorphism 
group of  :math:`\mathcal{P}` defined in section  
:ref:`aut_ploop_label`.
The group :math:`N_{0}` has a subgroup isomorphic to
:math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` generated by the
generators  :math:`x_\delta, \delta \in \mathcal{C}^*`
and :math:`x_\pi,  \pi \in {{\rm Aut}_{{\rm St}} \mathcal{P}}`.
The generators  :math:`x_\delta` generate the subgroup of diagonal
automorphisms of :math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}`. 


The group :math:`N_{0}` is generated by :math:`N_{x0}` and by 
another element :math:`\tau` of :math:`N_{0}` or order ``3``. The
group :math:`G_{x0}` is generated by :math:`N_{x0}` and by another
element  :math:`\xi` of :math:`G_{x0}` or order ``3``. The elements
:math:`x_\delta`, :math:`x_\pi`, :math:`x_e`, 
:math:`y_e`, :math:`z_e`,  :math:`\tau` and  :math:`\xi`  
generate the monster group  :math:`\mathbb{M}`.
In this  package we use the definitions of the generators 
in  :cite:`Seysen20`, which incorporate a modification of 
:cite:`Con85` made in :cite:`Iva09`. 
This leads to simpler relations in  :math:`N_{0}`.
The generators :math:`x_e`, :math:`y_e`, and :math:`z_e` in  
:cite:`Seysen20` correspond to the generators
:math:`x_e \cdot z_{-1}^{|e/4|}`, :math:`y_e \cdot x_{-1}^{|e/4|}`,
and  :math:`z_e \cdot y_{-1}^{|e/4|}` in :cite:`Con85`.
 

Representing elements of the monster group
..........................................

An element of the monster group :math:`\mathbb{M}` is represented 
as an instance of class ``MM`` in module ``mmgroup``.

Elements of the monster group are created by the constructor
of class |MM|. Usually, the constructor of class |MM| takes two
arguments ``tag`` and ``i``, where ``tag`` is a single small 
letter describing the type of the generator of :math:`\mathbb{M}`, 
and ``i`` is an integer describing the value of that generator. 
Alternatively, ``i`` may be an instance of the appropriate algebraic 
structure used for indexing a generator of a given type as 
indicated in the table below.


Implementation of the generators of the monster group
.....................................................


Math papers may use (at least) Latin or Greek letters for labelling
objects, but most programming languages are restricted to ASCII characters.
The following table shows how to create generating of the monster group
using the constructor of class |MM|:


.. table:: Construction of generating elements of the monster
  :widths: 10 90


  ===================  ==========================================================
  Element              Construction as an instance of class ``MM``, 
  ===================  ==========================================================
  :math:`x_\delta`     ``MM('d', delta)``, ``delta`` an instance of class
                       |Cocode|; 

                       ``MM('d', delta)`` returns 
                       ``MM('d', Cocode(delta))`` for ``0 <= delta < 0x1000``. 

  :math:`x_\pi`        ``MM(pi)``, ``pi`` an instance of class |AutPL|;

                       ``MM('d', delta) * MM('p', n)`` is equivalent to 
                       ``MM(AutPL(delta, n))`` 
                       for ``0 <= delta < 0x1000``, ``0 <= n < 244823040`` .

  :math:`x_e`          ``MM('x', e)``, ``x`` an  instance of class |PLoop|;

                       ``MM(('x', e))`` returns ``MM('x', PLoop(e))`` for
                       ``0 <= e < 0x2000``.

  :math:`y_e`          ``MM('y', e)``,  ``e`` as in case :math:`x_e`. 

  :math:`z_e`          ``MM('z', e)``,  ``e`` as in case :math:`x_e`. 

  :math:`\tau^e`       ``MM('t', e)``, exponent  ``e`` is an integer 
                       which is taken modulo ``3``.

  :math:`\xi^e`        ``MM('l', e)``,  exponent  ``e`` is an integer 
                       which is taken modulo ``3``.
  ===================  ==========================================================


More possibilities for constructing elements of an instance of class 
|MM| are given in the description of that class. Multiplication and 
exponentiation of group elements works a usual.  


Multiplication of elements of the monster group
...............................................

Internally, an element of :math:`\mathbb{M}` is represented as a word
of the generators given above. Multiplication with the ``*`` operator 
is a concatenation of such words, followed by a reduction step.
Multiplication of two reduced elements of the monster group (including
the reduction of the result) takes less than 50 milliseconds on the
author's computer. 

The reduction after a group operation is done by a new method that 
tracks pairs of perpendicular 2A axes, see :cite:`Seysen22`
for details.
"""
# References in the __docstr__ see file docs/source/references.bib


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



import sys
import os
import collections
import re
import warnings
from numbers import Integral
import numpy as np
from random import randint, sample


from mmgroup.generators import rand_get_seed
from mmgroup.generators import gen_rng_modp
from mmgroup.structures.parse_atoms import ihex, TaggedAtom
from mmgroup.structures.mm0_group import MM0
from mmgroup.structures.abstract_group import singleton
from mmgroup.structures.abstract_mm_group import AbstractMMGroup
from mmgroup.structures.parse_atoms import  AtomDict
from mmgroup.structures.construct_mm import load_group_name     
from mmgroup.structures.construct_mm import iter_mm       
from mmgroup.generators import gen_leech2_reduce_type4
from mmgroup.generators import mm_group_invert_word
from mmgroup.clifford12 import xsp2co1_check_word_g_x0 
from mmgroup.clifford12 import xsp2co1_reduce_word      
from mmgroup.clifford12 import chk_qstate12
from mmgroup.clifford12 import xsp2co1_rand_word_G_x0
from mmgroup.clifford12 import xsp2co1_rand_word_N_0
from mmgroup.mm import mm_vector



try:
    from mmgroup import mat24
    from mmgroup.mat24 import MAT24_ORDER, pow_ploop, ploop_theta
except (ImportError, ModuleNotFoundError):
    from mmgroup.dev.mat24.mat24_ref import Mat24  
    mat24 = Mat24  
    MAT24_ORDER = Mat24.MAT24_ORDER
    pow_ploop = Mat24.pow_ploop
    ploop_theta = Mat24.ploop_theta
    del Mat24


from mmgroup.structures.ploop import Cocode, PLoop
from mmgroup.structures.autpl import StdAutPlGroup, autpl_from_obj, AutPL

from mmgroup.generators import mm_group_mul_words






###########################################################################
# Import functions from module ``mmgroup.structures.mm_order`` on demand
###########################################################################


# Functions to be imported from module mmgroup.mm_order
check_mm_order = None
check_mm_equal = None
check_mm_half_order = None
check_mm_in_g_x0 = None
mm_reduce_M = None

def import_mm_order_functions():
    """Import functions from module ``mmgroup.mm_order``.

    We import these functions from module ``mmgroup.mm_order``
    on demand. This avoids an infinite recursion of imports.
    """
    global check_mm_order, check_mm_equal
    global check_mm_half_order, check_mm_in_g_x0
    global mm_reduce_M
    from mmgroup.structures.mm_order import check_mm_order
    from mmgroup.structures.mm_order import check_mm_equal 
    from mmgroup.structures.mm_order import check_mm_half_order
    from mmgroup.structures.mm_order import check_mm_in_g_x0
    from mmgroup.mm_reduce import mm_reduce_M 





###########################################################################
# Word class for the group MM
###########################################################################





class MM(MM0):
    r"""Models an element of the monster group :math:`\mathbb{M}`

    Let ``g1`` and ``g2`` be instances of class  |MM| representing 
    elements of the monster group.  Then
    ``g1 * g2``  means group multiplication, and ``g1 ** n`` means
    exponentiation of ``g1`` with the integer ``n``. ``g1 ** (-1)`` 
    is the inverse of ``g``. ``g1 / g2`` means ``g1 * g2 ** (-1)``.
    We have ``1 * g1 == g1 * 1 == g1`` and ``1 / g1 == g1 ** (-1)``.

    ``g1 ** g2`` means ``g2**(-1) * g1 * g2``.   

    Let ``V`` be a vector space that is a representation of ``MM``,
    see class |MMVector| for details. An instance ``g1`` of 
    class ``MM`` operates on the vector space  ``V`` by right 
    multiplication.  

    :var tag:
        In the simplest case, parameter ``tag`` is a string of length 
        ``1`` that determines the type of the atomic group element. 
        There are also some special cases for parameter ``tag `` as
        indicated below. If ``tag`` is not given or equal to ``1`` 
        then  the neutral element of the monster group is constructed.

     
    :var i:
        Parameter ``i`` is number of the atomic element of a given ``tag``.
        Depending on the tag, ``i`` may be the number of an element of one 
        of the structures |PLoop|, |Cocode|, or the number of an element of 
        the Mathieu  group ``M_24``, as explained in class |AutPL|.
        An element :math:`\pi` of the group |AutPL| is mapped to the
        element :math:`x_\pi` of the Monster group.
        
        The number ``i`` may also be an instance of the appropriate class
        |PLoop|, |Cocode|, or |AutPL|, as indicated in the table below.

    :return: An element of the monster group 
    :rtype:  an instance of class |MM|



    **Standard tags**

    The tags listed in the following tables are standard tags that can
    be used for creating generators (or some short words of generators)
    of the monster group.
 

    .. table:: Atomic elements of the Monster group
          :widths: 8 20 72

          +-------+-----------------+----------------------------------------+
          | Tag   | Number  ``i``   | Type of element                        |
          +=======+=================+========================================+
          |``'p'``| ``0-244823039`` | The automorphism ``AutPL(0, i)`` of    |
          |       |                 | the Parker loop, see below.            |
          +-------+-----------------+----------------------------------------+
          |``'d'``| ``0-0xfff``     | The diagonal automorphism ``Cocode(i)``|
          |       |                 | in |AutPL|.                            |
          +-------+-----------------+----------------------------------------+
          |``'x'``| ``0-0x1fff``    | The element :math:`x_e`,  with         |
          |       |                 | ``e = PLoop(i)``.                      |
          +-------+-----------------+----------------------------------------+
          |``'y'``| ``0-0x1fff``    | The element :math:`y_e`,               |
          |       |                 | ``e = PLoop(i)``;                      |
          |       |                 | similar to tag ``'x'``.                |
          +-------+-----------------+----------------------------------------+
          |``'z'``| ``0-0x1fff``    | The element :math:`z_e`,               |
          |       |                 | ``e = PLoop(i)``;                      |
          |       |                 | similar to tag ``'x'``.  We have       |
          |       |                 | :math:`x_e y_e z_e = y_e x_e z_e = 1`. |
          +-------+-----------------+----------------------------------------+
          |``'t'``| ``0-2``         | The element :math:`\tau^i`,            |
          +-------+-----------------+----------------------------------------+
          |``'l'``| ``0-2``         | The element :math:`\xi^i`,             |
          +-------+-----------------+----------------------------------------+
          |``'q'``| ``0-0x1ffffff`` | Describes an element of the subgroup   |
          |       |                 | :math:`Q_{x0}`. See remark below.      |
          +-------+-----------------+----------------------------------------+
          |``'c'``| ``0-0xffffff``  | A representative of a right coset of   |
          |       |                 | :math:`N_{x0}` in :math:`G_{x0}`.      |
          |       |                 | See remark below.                      |
          +-------+-----------------+----------------------------------------+


    Apart from the standard tags there are also some tags for
    special purposes discussed in the following sections.

    Remarks
          
    If ``i`` is the string ``'r'`` then a random element with the   
    given tag is generated.  If ``i`` is the string ``'n'`` then
    a random element with the given tag is generated, which is 
    different from the neutral element with a very high 
    probability.

    If ``tag == 'd'`` then  ``i = 'o'`` generates a random odd 
    and ``i = 'e'`` generates a  random even diagonal automorphism.
    In this case `i`` may also be an instance of class |Cocode|, 
    representing the diagonal automorphism correspnding to the
    given element of the Golay cocode.   
        
    If ``tag == 'p'`` then ``i`` may also be an instance of class 
    |AutPL|. Then the returned atom is the Parker loop automorphism
    given by that instance. If ``i`` is an integer then the Parker 
    loop  automorphism ``AutPL(0, i)`` is returned. This 
    automorphism is the standard rpresentative of the i-th element 
    of the  Mathieu group ``M_24`` in the automorphism group of the 
    Parker loop.

    If ``tag`` is ``'x'``,   ``'y'`` or ``'z'`` then ``i`` 
    may also be an instance of class |PLoop|, representing an 
    element of the Parker loop. 

    The exponent ``i`` for a tag ``'t'`` or ``'l'`` is reduced
    modulo ``3``. 

    The tag ``'q'`` is useful for encoding an element of the 
    subgroup :math:`Q_{x0}` of the monster. An atom with 
    tag ``'q'`` and index ``i``  encodes the element 
    :math:`x_d \cdot x_{\delta} \cdot x_{\theta(d)}`, 
    with :math:`d` = ``PLoop(i >> 12)`` :math:`\in \mathcal{P}``,   
    :math:`\delta` = ``Cocode(i & 0xfff)`` 
    :math:`\in \mathcal{C}^*`,  
    and   :math:`\theta: \mathcal{P} \rightarrow \mathcal{C}^*`
    the cocycle of the Parker loop.

    The tag ``'c'`` with index ``i`` encodes a representative
    of the right coset :math:`N_{x0} g, \,  g \in G_{x0}` of  
    :math:`N_{x0}` in :math:`G_{x0}` that maps the standard
    frame  :math:`\Omega` in the Leech lattice modulo 2 to 
    a type-4 vector given by the index ``i``.
    Here ``i`` encodes a vector in the Leech lattice modulo 2
    as in the description for tag  ``'q'``. That vector must
    be of type 4. The sign bit of that vector is ignored. 
    Here ``i`` may also be an instance of class |XLeech2|
    representing a type-4 vector. We warn the user that the
    index ``i`` may specify any element of the corresponding
    right coset :math:`N_{x0} g`; and that this element may 
    depend on the version of the package.


    **Generating a random element of the subgroup of the monster**
 
    If parameter ``tag`` is equal to the string ``'r'`` then parameter 
    ``'i'`` should be a string describing a subgroup of the 
    monster group. Then a uniform distributed element from that
    subgroup is generated. The follwing subgroups are recognized:

    .. table:: Subgroups of the Monster group recognized
          :widths: 15 85

       +----------------+---------------------------------------------+
       | Parameter ``i``| Subgroup                                    | 
       +----------------+---------------------------------------------+
       | ``'M'``        | The monster group itself                    |
       +----------------+---------------------------------------------+
       | ``'G_x0'``     | Subroup  ``'G_x0'`` generated by            |
       |                | :math:`x_d, x_\delta, y_d, x_\pi, \xi`      |
       +----------------+---------------------------------------------+
       | ``'N_0'``      | Subroup  ``'N_0'`` generated by             |
       |                | :math:`x_d, x_\delta, y_d, x_\pi, \tau`     |
       +----------------+---------------------------------------------+
       | ``'N_x0'``     | Subroup  ``'N_x0'`` generated by            |
       |                | :math:`x_d, x_\delta, y_d, x_\pi`           |
       +----------------+---------------------------------------------+
       | ``'N_0_e'``    | Subroup of ``'N_0'`` of index 2 generated   |
       |                | by :math:`x_d, x_\delta, y_d, x_\pi, \tau`, |
       |                | :math:`\delta` even                         |
       +----------------+---------------------------------------------+
       | ``'N_x0_e'``   | Subroup of ``'N_x0'`` of index 2 generated  |
       |                | by :math:`x_d, x_\delta, y_d, x_\pi`,       |
       |                | :math:`\delta` even                         |
       +----------------+---------------------------------------------+

    Subgroup ``'G_x0'`` has structure :math:`2^{1+24}.\mbox{Co}_1`, and
    subgroup ``'N_0'`` has structure :math:`2^{2+11+22}.\mbox{M}_{24}`.
    ``'N_x0'`` is a subgroup of ``'N_0'`` of index 3.

    If Parameter ``i`` is an positive integer then an element of the
    monster of shape :math:`g_0 t_1 g_1 t_2 g_2 ... t_i g_i` is created,
    where :math:`g_j` is a random element of  ``'G_x0'``, and :math:`t_j` 
    is  :math:`\tau^{\pm 1}`.


    **Generating an element of monster from its internal representation**

    If parameter ``tag`` is equal to the string ``'a'`` then parameter 
    ``'i'`` should be an array-like object containing the
    internal representation of an element of the monster group.
 
    Internally, an element of the monster group is represented
    as an array of unsigned 32-bit integers, where each entry
    of the array describes a generator. See section
    :ref:`header-mmgroup-generators-label` for details.


    **Special types accepted as tags**
  
    In the simplest case the argument ``tag`` in the constructor is
    a string of length ``1``. There are a few other types which are
    also accepted as tags. For such a special tag, the following
    parameter ``i`` in the constructor is ignored. 

    Especially, we may encode a word of standard generators as a 
    list of tuples as indicated in the table below.

    .. table:: Legal types as tags in the constructor of class ``MM``
      :widths: 25 75

      +---------------------+-------------------------------------------+
      | Object of type      | Evaluates to                              |
      +=====================+===========================================+
      | type ``list``       | A list represents the product of its      |
      |                     | entries in the given order. An entry      |
      |                     | ``MM(tag, i)`` can be abbreviated  to     |
      |                     | the tuple ``(tag, i)``.                   |
      +---------------------+-------------------------------------------+
      | class |PLoop|       | The Parker loop element :math:`d`         | 
      |                     | given by that object is mapped to         |
      |                     | :math:`x_d \in \mathbb{M}`.               |
      +---------------------+-------------------------------------------+
      | class |AutPL|       | The Parker loop automorphism :math:`\pi`  | 
      |                     | given by that object is mapped to         |
      |                     | :math:`x_\pi \in \mathbb{M}`.             |
      +---------------------+-------------------------------------------+
      | class |Cocode|      | The Golay cocode element :math:`\delta`   |
      |                     | given by that object is mapped to         |
      |                     | :math:`x_\delta \in \mathbb{M}`.          |
      +---------------------+-------------------------------------------+
      | class |XLeech2|     | The element corresponding to that element |
      |                     | of  :math:`Q_{x0}` is created.            |
      +---------------------+-------------------------------------------+
      | class |MM|          | A deep copy of the given element of       | 
      |                     | :math:`\mathbb{M}` is made.               |
      +---------------------+-------------------------------------------+
      | ``str`` of length   | For an instance ``g`` of class ``MM``     |
      | greater than ``1``  | we have ``MM(str(g)) == g``.              |
      |                     |                                           |
      |                     | So we can reread                          |
      |                     | printed elements of ``MM``.               |
      +---------------------+-------------------------------------------+


    """
    MIN_LEN = 128
    _MAGIC = randint(0, 0xffffffffffffffff)
    __slots__ =  "length", "_data", "reduced"
    def __init__(self,  tag = None, i = None, *args, **kwds):
        self.reduced = 0
        atoms = iter_mm(self.group, tag, i)
        self._data = np.fromiter(atoms, dtype = np.uint32) 
        self.length = len(self._data)
        self._extend(self.MIN_LEN)
        self.reduce()

    def _t_shape(self):
        l = []
        self.reduce()
        for d in self.mmdata:
            tag = (d >> 28) & 7
            if tag == 5: l.append('T')
            if tag == 6: l.append('x')
        return "<" + "".join(l) + ">"




###########################################################################
# The class representing the group MM
###########################################################################


REDUCE_MODE = 1

@singleton
class MMGroup(AbstractMMGroup):
    r"""An instance of this class models the monster group as an object

    :param: None

    :return: An instance of the monster group 
    :rtype:  an instance of class |MMGroup|

    Class |MMGroup| is implemented as a singleton. Calling the constructor
    of that class always returns the same instance of that class. 

    Elements of the monster group are implemented as instances of class
    ``MM``. 
    """
    ERR_REDUCE = "Reduction in monster group failed, error = %d"
    __instance = None
    word_type = MM
    group_name = "M"


    def __init__(self):
        """ TODO: Yet to be documented     


        """
        super(MMGroup, self).__init__()



    def reduce(self, g1, copy = False):
        l1 = g1.length
        if not g1.reduced:
            if not mm_reduce_M:
                import_mm_order_functions() 
            if copy:
                g2 = self.word_type()
                length = mm_reduce_M(g1._data, g1.length, 
                    REDUCE_MODE, g2._data)
                if not 0 <= length <= 128:
                    raise ValueError(self.ERR_REDUCE % length)
                g2.length = length
                g2.reduced = True
                return g2
            else:
                a = np.zeros(128, dtype = np.uint32)
                length = mm_reduce_M(g1._data, g1.length, 1, a)
                if not 0 <= length <= 128:
                    raise ValueError(self.ERR_REDUCE % length)
                g1._setdata(a[:length])
                g1.reduced = True
                return g1
        return self.copy(g1) if copy else g1

        
    def _imul(self, g1, g2):
        l1, l2 = g1.length, g2.length
        g1._extend(2*(l1 + l2) + 1)
        g1._data[l1 : l1 + l2] = g2._data[:l2]
        l1 += l2
        g1.length = l1
        g1.reduced = False
        g1.reduce()
        return g1

    def _invert(self, g1):
        w = self.word_type()
        w._setdata(np.flip(g1.mmdata) ^ 0x80000000)
        w.reduced = False
        return w

    def copy_word(self, g1):
        result = self.word_type()
        result._setdata(g1.mmdata)
        result.reduced = g1.reduced
        return result

    def _equal_words(self, g1, g2):
        if not check_mm_equal is None:
            return g1.group == g2.group and check_mm_equal(g1, g2)
        try:
            import_mm_order_functions()
            return g1.group == g2.group and check_mm_equal(g1, g2)
        except:
            if g1.group != g2.group:
                return False
            g1.reduce()
            g2.reduce()
            if (g1.mmdata == g2.mmdata).all():
                 return True
            raise ValueError("Don't know if monster group elements are equal")

    def str_word(self, g):
        try:
            g.reduce()
        except:
            pass
        return super(MMGroup, self).str_word(g)




StdMMGroup = MMGroup()
MM.group = StdMMGroup
load_group_name(StdMMGroup, "M")






"""Implement parity-adjusted Hadamard matrices

A parity-adjusted 2**n times 2**n Hadamard matrix M is defined as
follows:

Assmme that rows and columns are numbered with integers 0,...,2**n-1.
Then we put M[i,j] = -1**(parity(i & j) ^ (parity(i) & parity(j))), 
where operator '&' and '^' are defined on integers i, j as in C, and 
parity(i) returns the bit parity of i interpreted as a binary
number.

A standard Hadamard matrix H is defined by H[i,j] = -1**parity(i & j).

It turns out that all non-monomial blocks (except for one 3 times 3 
block) occuring in the representation of the Monster group are 
parity-adjusted 2**n Hadamard matrices of size 4, 6 or 64; or they 
are Kronecker products (i.e. tensor products) of two such matrices.

In this module we define the relevant parity-adjusted Hadamard 
matrices. Some of these matrices are taken from module
autogroup.NonMonomialOp_l.
"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import os
import sys
import collections
import numpy as np 


from mmgroup.tests.spaces.rep_aux import pm_mat_from_function
from mmgroup.tests.spaces.rep_aux import pm_diag_from_function
from mmgroup.tests.spaces.sparse_mm_space import NonMonomialOp_t
from mmgroup.tests.spaces.sparse_mm_space import NonMonomialOp_l

from mmgroup.bitfunctions import bitparity

########################################################################
# Create Hadamard matrices and some of their variation.
########################################################################


def identity_matrix(n):
    """Return the n times n identity matrix"""
    m = np.zeros((n,n), dtype=np.int32)
    for i in range(n): m[i,i] = 1
    return m

def hadamard_matrix(lg_n):
    """Return a 2**lg_n times 2**lg_n Hadamard matrix"""
    n = 1 << lg_n
    f = lambda i, j: bitparity(i & j) 
    return pm_mat_from_function(f, n)


def scal_prod_parity(i, j):
    """Return sign of entry i, j of parity-adjusted Hadamard matrix  

    Then such a matrix M has entries 

         M[i,j] = (-1) ** scal_prod_parity(i, j) .
    """
    return bitparity(i & j) ^ (bitparity(i) & bitparity(j))

def parity_hadamard_matrix(lg_n):
    """Return parity-adjusted Hadamard matrix of size 2**lg_n.

    The matrix is returned as a numpy array of shape (n, n) with
    n = 2**lg_n and of type np.int32. 
    """
    n = 1 << lg_n
    return pm_mat_from_function(scal_prod_parity, n)

# Next we collect the candidates for parity-adjusted Hadamard matrices

msym16 = NonMonomialOp_l.MSYM16
mdiag16 = NonMonomialOp_l.MDIAG16
mat_l_16 =  mdiag16 @ msym16 @ mdiag16 






"""Test C function dealing with Leech lattice vectors mod 3 of type 3

In this script we test functions dealing with vectors of type 3 in the 
Leech lattice modulo 3. These functions are implemented in file 
gen_leech3.c and available in the extension mmgroup.generators.

We use the terminology defined in
the document *The C interface of the mmgroup project*, 
section *Description of the mmgroup.generators extension*.
"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint, choices #, shuffle, sample
from functools import reduce
from operator import __or__
from numbers import Integral
from collections import defaultdict

import numpy as np
import pytest

from mmgroup import MM0, Cocode
from mmgroup.generators import gen_leech2_op_word
from mmgroup.generators import gen_leech2_map_std_subframe




#####################################################################
# Creating test elements of G_x0
#####################################################################

def G_x0_samples():
    testdata = [
        [],
        [('p', 1000)],
        [('l', 1)],
    ]
    for d in testdata:
        yield MM0(d).mmdata
    for i in range(100):
        yield MM0('r', 'G_x0').mmdata






#####################################################################
# Test mapping of type 3 Leech vectors modulo 3 to vectors modulo 2
#####################################################################



preimages = [0x800000] + [Cocode([0,i]).ord for i in range(1,24)]



@pytest.mark.gen_xi
def test_maps_std_subframe(verbose = 0):
    """Test the C function ``gen_leech2_map_std_subframe``"""
    print("Testing  C function gen_leech2_map_std_subframe")
    a = np.zeros(24, dtype = np.uint32)
    a_ref = np.zeros(24, dtype = np.uint32)
    BAD_MASK = -0x2000000
    for g in  G_x0_samples():
        if verbose:
           print("Input", MM0('a', g)) 
        for i, v in enumerate(preimages):
            a_ref[i] = gen_leech2_op_word(preimages[i], g, len(g))
            assert a_ref[i] & BAD_MASK == 0
        res = gen_leech2_map_std_subframe(g, len(g), a)
        assert res == len(g)
        ok = (a == a_ref).all()
        if verbose or not ok:
            print("Output obtained")
            print(a)
            print("Output expected")
            print(a_ref)
            if not ok:
               raise ValueError("Test failed")
    print("passed")
                 






from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import types

from random import randint

from operator import __xor__
from functools import reduce

def add4(a,b):
    """scalar or vector addition in GF(4)"""
    if type(b) == types.ListType:
         if type(a) == types.ListType:
              assert len(a) == len(b)
              return [a[i] ^ x for i,x in enumerate(b)]
    return a ^ b 


def mul4(a,b):
    """scalar or componentwise vector multiplication in GF(4)"""
    if type(b) == types.ListType:
         if type(a) == types.IntType:
              return [mul4(a,x) for x in b]
         if type(a) == types.ListType:
              assert len(a) == len(b)
              return [mul4(a[i],b[i]) for i in range(len(a))]
    if a * b: return 1 + (a+b-2) % 3
    return 0 


def conj4(a):
    """conjugates a number or vector in GF(4)"""
    if type(a) == types.ListType:
        return [conj4(x) for x in a]
    return [0,1,3,2][a] 
    


def tr4(a):
    """returns trace of a number or componentwise of a vector in GF(4)"""
    if type(a) == types.ListType:
        return [tr4(x) for x in a]
    return a >> 1


def parity(a):
    """returns parity, i.e. sum of entries, of a vector in GF(2) or GF(4)"""
    return reduce(__xor__, a, 0)




def exp4(i):
    """returns \eps**i in GF(4), coded as 2,3,1 for i=0,1,2"""
    return 1 + i%3


def hermite4(a,b):
    """returns tr(a*b**2) for scalars or componentwise for vectors in GF(4)"""
    if type(b) == types.ListType:
         if type(a) == types.IntType:
              return [ hermite4(a,x) for x in b ]
         if type(a) == types.ListType:
              assert len(a) == len(b)
              return [ hermite4(a[i],x) for i,x in enumerate(b) ]
    if a*b: return int(a != b)
    return 0




def quadrilinear_hexacode(a,b,c,d):
   """reuturns scalar product (  tr(a*b**2),  tr(c*d**2) ) """
   h1 = hermite4(a,b)
   h2 = hermite4(c,d)
   return sum([ h1[i] * h2[i] for i in range(6)]) & 1
   

BASIS = [
  [1,0,0,1,3,2],
  [0,1,0,1,2,3],
  [0,0,1,1,1,1]
]


def hexacode_basis(i,j=1):
   """returns j times the i-th hexacode basis vector"""
   return mul4(j, BASIS[i])


def hexacode_random():
   """returns a random hexacode vector"""
   a = [0]*6
   for i in range(3):
      a = add4(a,hexacode_basis(i,randint(0,3)))
   return a


def iter_hexacode():
    """iterator over all 64 hexacode vectors"""
    def mul_basis(i):
        return [ hexacode_basis(i,j) for j in range(4) ]
    for x in mul_basis(0):
        for y in mul_basis(1):
            for z in mul_basis(2):
                 yield(add4(add4(x,y),z))

def iter_hexacode_basis():
    """iterator over nonzero multiples of hexacode basis vectors"""
    for i in range(3):
        for j in range(1,4):
            yield hexacode_basis(i,j)

def weight(a):
    """returns the weight of a vector. i.e. the No of nonzero entries"""
    return sum(map(bool,a))


def cartesian(*lists):
    """returns the Cartesian product of the argument, each argument a list"""
    if len(lists) == 0:
        yield []
        return
    hd, tl = lists[0], lists[1:]
    for h in hd:
        for l in cartesian(*tl):
            yield [h] + l





def cocode_product(a,b):
    c = conj4(mul4(a,b))
    for i in range(3):
        c = add4(c, hexacode_basis(i,c[i]) )
    for i in range(3): assert c[i] == 0
    return c 


def reduce_to_cocode(a):
    c = a[:]
    for i in range(3):
        c = add4(c, hexacode_basis(i,c[i]) )
    for i in range(3): assert c[i] == 0
    return c 

def is_hexacode_word(a):
    return reduce_to_cocode(a) == [0] * 6
  
LIGHT_COCODE_WORDS = None

def light_cocode_words():
    """returns dictionary mapping cocode words to lists of light cocode words

    Here the keys in the dictionary are the possible return values of
    function reduce_to_cocode().

    The light cocode words are the words of weight <= 2
    """
    ## TODO: cleanup and document this !!!!!!!!!!!!!!!!
    global LIGHT_COCODE_WORDS
    if not LIGHT_COCODE_WORDS is None:
        return LIGHT_COCODE_WORDS
    LIGHT_COCODE_WORDS = {}
    def add_word(cocode_word, light_word):
        cw = tuple(cocode_word)
        if not cw in LIGHT_COCODE_WORDS.keys():
             LIGHT_COCODE_WORDS[cw] = [ light_word[:] ]
        else:
             LIGHT_COCODE_WORDS[cw] = ( LIGHT_COCODE_WORDS[cw] + [light_word[:]] )
    add_word([0]*6, [0]*6) 
    for i in range(6):
         for wi in range(1,4):
             lw = [0]*i + [wi] +  [0]*(5-i)
             cw = reduce_to_cocode(lw)
             add_word(cw, lw)
             for j in range(i+1,6):
                 for wj in range(1,4):
                     lw2 = lw[:]
                     lw2[j] = wj
                     cw = reduce_to_cocode(lw2)
                     add_word(cw, lw2)
    return LIGHT_COCODE_WORDS 


def decode_cocode(a):
    """returns list of equivalent words to a of length <= 2 modulo the hexacode"""
    cw = tuple(reduce_to_cocode(a))  
    return  light_cocode_words()[cw]  

def codeword_type(a):
    """returns a tuple describing the type of a vector a in GF(4)^6.

    The 1st entry is the weight of the vector a. Other entries are present
    if a is no hexacode word. Then the other entries are the (sorted) tuple 
    of the weights of all hexacode words that have minimal  distance to a.

    """
    ## Baustelle: cleanup and document this !!!!!!!!!!!!!!!!
    wa = weight(a)
    wd = decode_cocode(a)
    if weight(wd[0]) == 0: return (wa,)
    cwd = [ add4(a,b) for b in wd]
    wd1 = sorted(map(weight, cwd))
    return tuple( [wa]  +  wd1 )

            


subword4_list = None

def subwords_4():
    """return pattern list of all words of weight 4"""
    global subword4_list
    if subword4_list is None:
         subword4_list = []
         for i in range(6):
             for j in range(i+1,6):
                 subword4_list.append( 
                     [3]*(i) + [0] + [3]*(j-i-1) + [0] + [3]*(6-j-1) )
    return subword4_list

def no_hexacode_subwords(a):
    """Returns True if no subword of a is a hexacode word

    subwords are obtained by replacing entries of a by zero.
    """
    if is_hexacode_word(a):
        return False
    for y in subwords_4():
        b = [a[i] & y[i] for i in range(6)]
        if is_hexacode_word(b):
            return False
    return True
 

def iter_content(length=6, content = [1,2,3]):
    """Iterates over all lists of given length with given content"""
    if length == 0:
        yield []
    else:
        for l in iter_content(length-1, content):
            for j in content:
                 yield l + [j]


    
  


def trace3(a,b,c):
    """return trace(a*b*c), a,b,c \in GF(4)"""
    return mul4(a,mul4(b,c)) >> 1

def trilinear(a,b,c):
    
    y = 0
    for i in range(6):
        y ^= trace3(a[i],b[i],c[i])
    return y 





###############################################################################




def test_hermite4():
    for a in range(4):
        for b in range(4):
            assert hermite4(a,b) == tr4(mul4(a, conj4(b))) 


def test_quadrilinear_nontrivial():
    print( "\nTest quadrilinear form on QA random hexacode vectors .." )
    for i in range(100):
        a = hexacode_random()
        b = hexacode_random()
        c = hexacode_random()
        d = hexacode_random()
        x = quadrilinear_hexacode(a,b,c,d) 
        y = quadrilinear_hexacode(a,c,b,d) 
        z = quadrilinear_hexacode(a,d,c,b)
        if x == y == z == 0: continue  
        print( a, b, c, d )
        print( x,y,z )
        print( "quadrilinear form QA is not trivial\n" )
        return



def test_quadrilinear(a,b,c,d):
    A = quadrilinear_hexacode(a,b,c,d) 
    B = quadrilinear_hexacode(a,c,b,d) 
    C = quadrilinear_hexacode(a,d,c,b) 
    assert A ^ B ^ C == 0 


def test_quadrilinear_on_basis():
    for a in iter_hexacode_basis():
        for  b in iter_hexacode_basis():
             for  c in iter_hexacode_basis():
                 for  d in iter_hexacode_basis():
                     test_quadrilinear(a,b,c,d)
    print( "QA(a,b,c,d) + QA(a,c,b,d) + QA(a,d,c,b) = 0""" )
    for a in iter_hexacode():
        for b in iter_hexacode():
            assert parity( hermite4(a,b) ) == 0   
    print( "QA(a,b,a,b) = 0" )




def test_trilinear_form():
    print( "put A(a,b,c) = sum_{i} tr(a[i] b[i] c[i])" )
    for a in iter_hexacode_basis(): 
        for b in iter_hexacode_basis():
            for c in iter_hexacode_basis():
                if  trilinear(a,b,c):
                    print( "A(",a,',',b,',',c,") = 1" )
                    print( "trilinear form A : hexacode -> GF(2) is not trivial" )
                    return

def test_self_dual():
    for a in iter_hexacode(): 
        for b in iter_hexacode():
            assert sum(hermite4(a,b)) & 1 == 0
    print( "Hexacode is self-dual, i.e. hermite4(a,b) = 0 for codewords a,b" )


def hermite4_cocode_nonzero():
   print( "Example where hermite4(a,b) = (..,tr(a_i*b_i^2),..) is zero" )
   print( "and cocode product A(a,b,.) is not zero" )
   null = [0]*6
   for a in iter_hexacode(): 
      for b in iter_hexacode():
          c = cocode_product(a,b)
          if  c != null:
              try:  
                  assert hermite4(a,b) != null, (a,b,c, hermite4(a,b) )
              except: 
                  print( "hermite4(",a,',',b,") = 0" )
                  print( "cocode product is",  conj4(mul4(a,b)) )  
                  return




def print_no_hexacode_subwords():
    n = 0
    for a in iter_content(6,  content = [1,2] ):
        if  no_hexacode_subwords(a) and sum(a) == 8:
            n += 1
    print ("\nThere are", n, "words containing four times 1 and twice 2")

    L2 = []
    print( " such that no subword is a hexacode word" )
    for x0 in iter_content(4,  content = [2,3] ):
       x = [1,1] + x0
       if sum(x) == 12 and  no_hexacode_subwords(x):
           L2.append(x)
    print( "For the following words no subword is a hexacode word:" )
    print( L2 )




def hexacode_types():
    # We go thru the Cartesian product [0,1,2,3] x ... x [0,1,2,3] with 6 factors"""
    all_words = list(cartesian( *([[0,1,2,3]]*6)))
    print("""\nDetermine types of all words of GF(4)^6 with respect to the hexacode:

The 1st entry is the weight of the vector v. Other entries are present
if v is no hexacode word. Then the other entries are the (sorted) tuple 
of the weights of all hexacode words that have minimal distance to v.""")
    types = {}
    samples = {}
    for a in all_words:
        newtype = codeword_type(a)
        if not newtype in types:
             types[newtype] =0
             samples[newtype] = (a, decode_cocode(a))
        types[newtype] += 1
    n = 0
    for k in sorted(types.keys()):
         if  k[0] != n: print( '' )
         print( str(k) + ':' + str(types[k]) +',', end=' ' )
         n =k[0] 
    return types
 
    




#######################################################################



def selftest():
    s ="""Let H be the Hexacode over GF(4)**6, QA: H**4 -> GF(2) the quadrilinear
form given by QA(a,b,c,d) = (  tr(a*b**2),  tr(c*d**2) ) with (.,.) the 
scalar product on GF(2)**6 and tr(a*b**2) computed componentwise on the 
hexcode.  Note that  tr(a*b**2) is in GF(2) for a,b in GF(4),
"""

    print( s )

    print("Hexacode basis vectors and their multiples, with '2' a generator of GF(4):")
    for i in range(3):
       for j in range(1,4):
         print ( hexacode_basis(i,j), ',' , end=' ' )
       print ("")


    test_quadrilinear_nontrivial()

    test_quadrilinear_on_basis()

    print ("")

    test_trilinear_form()
    print ("")
                   
    test_hermite4()

    hermite4_cocode_nonzero()

    test_self_dual()


    print_no_hexacode_subwords()


    hexacode_types()


if __name__ == "__main__":
    selftest()




r"""Module ``mat24tables`` contains the class ``Mat24Tables``

Class ``Mat24Tables`` is a base for class ``Mat24`` in module
``mmgroup.dev.mat24.mat24_ref``. It contains the basis of the Golay code 
and of (a set of representatives of) the Golay cocode. It also contains 
tables for fast conversion of a vector in ``V = GF(2)**24`` from the 
standard to the internal representation and vice versa. That class also 
contains a table of the syndromes of all ``2048`` odd elements of the 
cocode. The ``759`` octads, i.e. Golay code word of length ``8`` are 
numbered from ``0`` to ``758``. Class ``Mat24Tables`` provides tables
for computing the number of a octad from a Golay code word representing
an octad and vice versa. 

Class ``Mat24Tables`` also contains python versions of some C functions
in file ``mat24_functions.c`` using these tables.

"""
from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

# python 3 string formatting see:
# https://pyformat.info/


import types
import sys
import re
import os
from random import randint
from operator import __or__, __xor__, __and__
from functools import reduce

import numpy
from numpy import array, zeros, uint8, uint16, uint32

from mmgroup.bitfunctions import pivot_binary_high, bit_mat_transpose
from mmgroup.bitfunctions import bit_mat_inverse, bit_mat_basis_spanned
from mmgroup.bitfunctions import reverse24, lin_table, bw24
from mmgroup.bitfunctions import iter_bitweight, lmap


from mmgroup.dev.mat24.mat24aux import Lsbit24Function


############################################################################
############################################################################
# Lexicographical basis of the Golay code
############################################################################
############################################################################

lex_golay_code_basis = None

def make_lex_golay_code_basis():
    r"""Return lexicographical basis of the Golay code.

    This basis is computed from the definition of the Golay code as a 
    lexicographic code, see :ref:`CCN+85`, Chapter. 11,  section 17.  

    This computation takes a long time. So we better save the result
    in the global variable 'lex_golay_code_basis'.

    This function is mainly of historical interest. It has been run
    once for a proof that our definition of the Golay code agrees with 
    that in :ref:`CCN+85'. 
    """ 
    global lex_golay_code_basis 
    if not lex_golay_code_basis is None:
        return lex_golay_code_basis
 
    print("Computing the lexicographical basis of the Golay code")
    bitweights = numpy.zeros(1 << 24, dtype = numpy.uint8)
    for i in range(24):
        bitweights[1<<i:2<<i] = 1 + bitweights[0:1<<i]
    # basis of already generated part of the Golay code 
    basis = []
    # codes[:len_code] is the list of all code worde generated from basis
    codes = numpy.zeros(4096, dtype = numpy.uint32)
    len_codes = 1
    for d in range(7,24):
        # For 8<=d1<=24 loop through all d1-bit integers i of bit weight 8
        for i in range(1<<d, 2<<d):
            if bitweights[i] == 8:
                bad = False
                # set 'bad' if an element of codes[] has distance < 8 from i
                for c in codes[:len_codes]:
                    bad = bitweights[c ^ i] < 8
                    if bad: break
                # If not bad: append i to basis and update codes
                if not bad:
                    basis.append(i)
                    if d < 24:
                        for j in range(len_codes):  
                            codes[len_codes+j] = i ^ codes[j]
                        len_codes <<= 1
                    break
    lex_golay_code_basis = basis
    print("Lexicographical basis of the Golay code has been computed")
    del bitweights
    del codes
    return basis

############################################################################
############################################################################
# The hexacode (used for making up a basis of the Golay code)
############################################################################
############################################################################


class HexacodeToGolay(object):
    r"""This class deals with the mapping from the Hexacode to the MOG
    
    This mapping is described in :ref:`CCN+85`, ch. 11, section 5. Rows of 
    the MOG are  numbered from 0 to 5 (left to right). Columns of the MOG 
    are numbered from 0 to 3 (top to bottom). MOG entries are represented 
    as integers with bit 4*i+j (of valence 2**(4*i+j)) corresponding to the 
    MOG entry in row i, column j::


                         +------------------------+
                         |  0   4   8  12  16  20 |                 
                         |  1   5   9  13  17  21 |                 
                         |  2   6  10  14  18  22 |                 
                         |  3   7  11  15  19  23 |                 
                         +------------------------+

    A hexacode word (or more precisely, an element of GF(4)^6) is 
    represented as a list of six entries of value 0,...,3, so that 0  
    and 1 are in the subfield GF(2) of GF(4) and an entry in column j
    has 'score' j, see  :ref:`CCN+85`, ch. 11, section 5.
    """
    HEXA_ENCODE = [0x0,0xc,0xa,0x6] 
         # Encoding of an even interpretation with score j:
         # HEXA_ENCODE[j] << (4*i) is the unique MOG vector with an 
         # even number of nonzero entries in column i (none in any other
         # column or in row 0) and with 'score' j in column i.

    HBASIS =  [
        [1,0,0,1,3,2], [0,1,0,1,2,3], [0,0,1,1,1,1],
        [2,0,0,2,1,3], [0,2,0,2,3,1], [0,0,2,2,2,2],
    ] 
         # This the GF(2) basis of the hexacode in [SPLG], ch. 11.

    HISTORICAL_REVERSED_BASIS = [
        0xff, 0xf0f, 0x3333, 0x5555, 0x9669, 0x30356, 
        0x50563, 0x9063a, 0x111178, 0x21121d, 0x41144e, 0x811724
    ]
         # Golay code basis taken from an older implementation for 
         # compatibility checks.


    basis_ = None
         # store basis here after the 1st call to cls.basis()

    @classmethod
    def show_basis(cls,basis):
        """display a basis of the Golay code or cocode"""
        for x in basis:
            print( "%06x" % x, end = " ")
        print( "" )

    @classmethod
    def hexacode_vector_to_mog(cls, h):
        """Maps a hexacode word h to a Golay code word.

        Entry h[i] is mapped to an even interpretation of MOG column i, 
        with score h[i], and the top bit of column i being zero, 
        see [SPLG], ch.11, section 5.         
        """  
        l = [ cls.HEXA_ENCODE[x]<<(4*i) for i,x in enumerate(h) ]
        return sum(l)

    @classmethod
    def mog_std_octad(cls, i):
        """Returns the MOG octad with ones in row 0 and column i

        The entry at the intersection of row 0 and column i is zero.
        """  
        return 0x111111 ^ (0xf << (4*i))

    @classmethod
    def historical_check(cls, GolayBasis, generate = False):
        """Checks the Golay code basis against basis a stored older basis

        The basis is checked against basis cls.HISTORICAL_REVERSED_BASIS.
        If 'generate' is set, the basis cls.HISTORICAL_REVERSED_BASIS
        is recomputed by function  make_lex_golay_code_basis().
        """
        ref = lmap(reverse24, cls.HISTORICAL_REVERSED_BASIS)  
        ref = pivot_binary_high(ref)[0] 
        assert  GolayBasis == ref
        if generate:
            assert HISTORICAL_REVERSED_BASIS ==  make_lex_golay_code_basis()


    @classmethod
    def any_golay_basis(cls, verbose = False):
        """Returns a basis of the Golay code.

        This basis is directly calculated from the hexacode.
        """
        basis = list(map(cls.hexacode_vector_to_mog, cls.HBASIS))
        basis += [cls.mog_std_octad(i) for i in range(6)]
        basis = pivot_binary_high(basis)[0]
        cls.historical_check(basis)
        return basis


    @classmethod
    def check_cocode_basis(cls, cocode_basis):
        """Checks correcteness of a basis of a transversal of the Golay cocode"""
        assert len(cocode_basis) == 12, cocode_basis
        bit_mat_inverse( cls.any_golay_basis() + cocode_basis )
            # this inverse must exist!!

    @classmethod
    def better_cocode_basis(cls, verbose=False):
        """Returns a good basis of a traversal of the Golay cocode.

        The asserted properties of the cocode basis are as in the
        main comment of module mat24.py.
        """
        def e(i,a):
            v = [0]*6
            v[i] = a
            return v
        hexa = lmap(e, [2,2,1,1,0,0], [2,1,2,1,2,1])
        #basis = [0x111111,0x011111,0x110000,0x101000,0x100100,0x100010,]
        basis = [0x111111,0x1,0x110,0x1010,0x10010,0x100010,]
        basis +=  lmap(cls.hexacode_vector_to_mog, hexa)
        basis = basis[2:] + basis[:2]        
        cls.check_cocode_basis(basis)
        return basis



    @classmethod
    def code_basis_from_cocode_basis(cls, cocode_basis):
        """return orthogonal Golay code basis of the given cocode basis"""
        a = bit_mat_inverse(cls.any_golay_basis() + cocode_basis)
            # this inverse must exist!!
        res =  bit_mat_transpose(a)[12:]
        assert cls.any_golay_basis() == pivot_binary_high(res)[0]
        return res

    @classmethod
    def better_code_basis(cls):
        """returns a good basis of the Golay code

        The asserted properties of the Golay code basis are as in the
        main comment of module mat24.py.
        """
        if getattr(cls, "_BETTER_CODE_BASIS", None) is None:
            cls._BETTER_CODE_BASIS = cls.code_basis_from_cocode_basis(
                                                  cls.better_cocode_basis())
        return cls._BETTER_CODE_BASIS
        
    @classmethod
    def basis(cls):
        """Return basis for the GF(2)**24 to be used in the Golay code

        basis [0:12] is the basis for a transversal of the Golay code
        given by method better_cocode_basis().
        basis [12:24] is the basis of the Golay code in GF(2)**24 
        given by method better_code_basis().
        """
        if cls.basis_ is None:
            basis =  cls.better_code_basis()
            cobasis =  cls.better_cocode_basis()
            cls.basis_ = numpy.array(cobasis + basis, dtype = uint32)
        return cls.basis_

   
############################################################################
############################################################################
# Conversion tables for the Golay code
############################################################################
############################################################################


def encoding_tables(basis):
    table0 = bit_mat_basis_spanned(basis[0:8], uint32)
    table1 = bit_mat_basis_spanned(basis[8:16], uint32)
    table2 = bit_mat_basis_spanned(basis[16:], uint32)
    return table0, table1, table2


def make_syndrome_table(recip_basis):
     """Generates a table for fast calulation of syndrome of a bit vector

     Table input is a 11-bit number representing an odd cocode word
     ``c``. Since basis vector 11 is the only vector with odd parity, 
     exactly  the cocode words with number ``c  >= 0x800`` have odd 
     parity. In table entry ``c & 0x7ff``  we list the unique cocode 
     word of weight 1 or 3 equivalent to the cocode word with number
     ``c``in the following format:

              i   +  (j << 5)  +  (k << 10) ,

     representing the word with ones at position i, j and k. If that
     cocode word has weight 1, we put j = k = 24.

     We set bit 15 of entry ``e`` in the table to one iff the (even)
     cocode word with number ``e`` has bit weight 2.

     Input recip_basis is the reciprocal basis of the choosen basis
     for the Golay code. (Lower 12 basis vectors must span a 
     transversal of the Golay code. All but the least of these 
     lower basis vector must have even parity.)
     """ 
     syndrome_table = zeros(0x800, dtype=uint16)
     C1 = (24 << 5) | (24 << 10) 
     for i in range(24):
         assert recip_basis[i] & 0x800
     rb = recip_basis & 0x7ff
     for i in range(24):
         b_i = rb[i]
         syndrome_table[b_i] ^=  i | C1 
         for j in range(i+1,24):
             b_j = b_i ^ rb[j]
             syndrome_table[b_j] ^= 0x8000
             for k in range(j+1,24):
                 b_k = b_j ^ rb[k]
                 syndrome_table[b_k] ^= i | (j<<5) | (k<<10)  
     assert min(syndrome_table & 0x7fff) > 0
     return syndrome_table
                 



def make_octad_tables(basis):
    """Return tables for encoding an decoding octads.

    Octads are numbered in the lexicographical order as given by the 
    numbers of the corresponding Golay codewords in 'gcode' 
    representation. 

    The function returns a triple 

       (oct_enc_table, oct_dec_table, oct_enc_offset)

    Given octad o, the correspondig Golay code word in gcode
    representation is oct_dec_table(o), for 0 <= 0 < 759.

    Given an octad (or a complement of an octad) v in 'gcode' 
    representation, the number of the corresponding octad is:

       (oct_enc_table[v1] >> 1) + 3 * (v1 >> 3) - oct_enc_offset ,

    where  v1 = v & 0x7ff.

    The vector v is a (possibly complemeted) octad if the following 
    condition holds:

       oct_enc_table[v1] < 255  .

    It is not a complemented octad if in addition we have:

       (v >> 12) & oct_enc_table[v1] & 1 == 0

    """
    codewords = lin_table(basis[:11])
    oct_dec_table = numpy.zeros(759, dtype = uint16)
    octad = 0
    d = {}
    w = {}
    for gcode, vector in enumerate(codewords[:2048]):
        weight = bw24(vector)
        if weight in [8, 16]:
            oct_dec_table[octad] = gcode + ((weight & 16) << 7)
            d[gcode] = octad - 3 * (gcode >> 3)
            w[gcode] = weight >> 4
            octad += 1
    assert octad == 759
    d_min, d_max = min(d.values()), max(d.values())
    assert d_min <= 0
    assert d_max - d_min < 127
    oct_enc_table = numpy.full(2048, 255, dtype = uint8)
    for gcode, dict_value in d.items():
        new_value = dict_value - d_min
        oct_enc_table[gcode] = w[gcode] + 2 * new_value
    return oct_enc_table, oct_dec_table, -d_min


class Mat24Tables(object):
    basis = HexacodeToGolay.basis()
    recip_basis = numpy.array(bit_mat_inverse(basis), dtype=uint32) 
    enc_table0, enc_table1, enc_table2 = encoding_tables(recip_basis)
    dec_table0, dec_table1, dec_table2 = encoding_tables(basis)
    syndrome_table = make_syndrome_table(recip_basis)
    #oct_enc_table, oct_dec_table, oct_index_table = make_octad_tables_old( 
    #    basis[12:] )
    oct_enc_table, oct_dec_table, oct_enc_offset = make_octad_tables(
        basis[12:] )

    ###########################################################################
    # Some general bit operations
    ###########################################################################

    @staticmethod
    def lsbit24(v1):
        """Return position of least significant bit of the integer v1

        In case v1 & 0xffffff == 0 we return 24.
        """
        return Lsbit24Function.compute(v1)
        
    @staticmethod
    def bw24(v1):
        """Return the bit weight of the lowest 24 bits of v1"""
        v1 = (v1 & 0x555555) + ((v1 & 0xaaaaaa) >> 1)
        v1 = (v1 & 0x333333) + ((v1 & 0xcccccc) >> 2)
        v1 = (v1 + (v1 >> 4)) & 0xf0f0f
        return (v1 + (v1 >> 8) + (v1 >> 16)) & 0x1f 

    ###########################################################################
    # Basis of Golay cocode and code
    ###########################################################################

    @classmethod
    def golay_basis(cls):
        """Return basis of Golay cocode followed by basis of Golay code

        The basis is returned as a list of 12 + 12 integers
        """
        return [cls.basis[i] for i in range(24)]


    @classmethod
    def basis_weights_8(cls):
        """Return weights of basis of Golay code

        Let ``w[i]`` be the bit weight of the ``i``-th basis
        vector of the Golay code.
 
        The function returns a integer ``v`` such that bit ``i``  
        of ``v`` is equal to ``(w[i] >> 3) & 1``.
        """
        w = [cls.bw24(cls.basis[i + 12]) for i in range(12)]
        assert reduce(__or__, w) & 7 == 0
        bits = [(b >> 3) & 1 for b in w]
        return sum((b << i for i, b in enumerate(bits)))



    ###########################################################################
    # Conversion between bit vectors of GF(2)**24
    ###########################################################################
 
    @staticmethod
    def vect_to_bit_list(v1):
        """Return lists of positions of 1-bits and 0-bits of v1 

        Let w be the bit weight of v1 & 0xffffff, i.e. number n of bits 
        of v1 at positions < 24 equal  to one. Then the ordered bit
        positions where the corrsponding bit of v1 is 1 are stored in
        a list  l1.
       
        Then v1 & 0xffffff has  24 - w  zero bits. The ordered list of 
        the positions of these zero bits is stored in l0. 

        The function returns the pair len(l1), l1 + l0.
        This return value is for compatibility with the corresponding
        C function.
        """
        l1 = [ i for i in range(24) if v1 & (1 << i) ]
        l0 = [ i for i in range(24) if not v1 & (1 << i) ]
        return len(l1), l1 + l0


    @staticmethod
    def extract_b24(v1, u_mask):
        """Extract the bits of 24-bit vector v1 given by the mask u_mask
   
        If u_mask has bits equal to one at positions i_0, i_1, ..., i_k
        (in ascending order) then the bit of v1 at position i_j is copied 
        to the bit at position j of the return value for j = 0,...,k.
        """
        l = [(v1 >> i) & 1 for i in range(24) if (u_mask >> i) & 1] 
        return sum(x << i for i, x in enumerate(l))

    @staticmethod
    def spread_b24(v1, u_mask):
        """Spread the bits of 24-bit vector v1 according to the mask u_mask

        If u_mask has bits equal to one at positions i_0, i_1, ..., i_k,
        (in ascending order) then the bit of v1 at position i is copied 
        to the bit at position i_j of the return value for j = 0,...,k.
        """
        l = [i for i in range(24) if (u_mask >> i) & 1] 
        return sum( ((v1 >> i) & 1) << x for i, x in enumerate(l) )




    ############################################################################
    # Conversion between representations of GF(2)**24, Golay code, etc.
    ############################################################################


    @classmethod
    def vect_to_vintern(self, v1):
        return  int(    self.enc_table0[v1 & 0xff]
                   ^ self.enc_table1[(v1 >> 8) & 0xff]
                   ^ self.enc_table2[(v1 >> 16) & 0xff]  )


    @classmethod
    def vintern_to_vect(self, v1 ):
        return  int(    self.dec_table0[v1 & 0xff]
                   ^ self.dec_table1[(v1 >> 8) & 0xff]
                   ^ self.dec_table2[(v1 >> 16) & 0xff]  )

    @classmethod
    def vect_to_cocode(self, v1 ):
        return self.vect_to_vintern(v1) & 0xfff

    @classmethod
    def gcode_to_vect(self, v1):
        return  int(   self.dec_table1[(v1 << 4) & 0xf0]
                   ^ self.dec_table2[(v1 >> 4) & 0xff]  )

    @classmethod
    def cocode_to_vect(self, c1):
        return self.vintern_to_vect(c1  & 0xfff)

    @classmethod
    def vect_to_gcode(self, v1):
        cn = self.vect_to_vintern(v1)
        assert cn & 0xfff == 0, "not a Golay code word"
        return cn >> 12


    @classmethod
    def vect_to_octad(self, v1):
        return self.gcode_to_octad(self.vect_to_gcode(v1))


    @classmethod
    def gcode_to_octad(self, v1, u_strict = 1):
       y = int(self.oct_enc_table[v1 & 0x7ff])
       if ((v1 >> 11) ^ y) & u_strict & 1 or y >= 254:
           raise ValueError("Golay code vector is not an octad")
       v1 &= 0x7ff
       return (y >> 1) + 3 * (v1 >> 3) - self.oct_enc_offset 


    @classmethod
    def octad_to_gcode(self, u_octad):
        assert 0 <= u_octad < 759, "not a Golay code octad"
        return self.oct_dec_table[u_octad] & 0xfff

    @classmethod
    def octad_to_vect(self,u_octad):
        return self.gcode_to_vect(self.octad_to_gcode(u_octad))



    ###########################################################################
    # Golay code syndromes and weights
    ###########################################################################


    @classmethod
    def cocode_syndrome(cls, c1, u_tetrad=24):
        """Return equivalent Golay cocode to 'c1' of length at most four.

        Same as mat24_syndrome(mat24_cocode_to_vect(c1, u_tetrad)).  
        """
        if not 0 <= u_tetrad <= 24:
            raise ValueError("Bad argument in function cocode_syndrome()")
        bad = (u_tetrad >= 24) & ((c1 >> 11) + 1) & 1
        u_tetrad -= (u_tetrad + 8) >> 5      # change u_tetrad == 24 to 23
        y = - (((c1 >> 11) + 1) & 1)         # y = 0 if c1 is odd else -1
        c1 ^= cls.recip_basis[u_tetrad] & y  # make c1 odd, if c1 is odd:
        y  &=   1 << u_tetrad                #     set bit u_tetrad in y
        syn = cls.syndrome_table[ c1 & 0x7ff ]
        syn = (1 << (syn & 31)) | (1 << ((syn>>5) & 31)) | (1 << ((syn>>10) & 31))
        # bit 24 in syn is also set if syn has weight 1
        bad &= not bool(syn & (y | 0x1000000)) 
        if bad:
            raise ValueError("Golay code syndrome is not unique")
        syn ^= y
        return syn & 0xffffff


    @classmethod
    def syndrome(cls, v1, u_tetrad=24):
        """Return equivalent Golay cocode word to 'v1' of length at most four.

        If that cocode word has length four, 0 <= 'u_tetrad' < 24 is a bit 
        position where the (not uniquely defined) eqivalent cocode word will 
        have a 'one' bit. Otherwise, that cocode word is uniquely defined, and 
        'tetrad' is ignored. in any case 0 <= 'u_tetrad' <= 24 is required. 
        Here 24 means don't care.  Thus u_tetrad = Mat24.lsbit24(x) is legal 
        for any x. 
        """
        return cls.cocode_syndrome(cls.vect_to_vintern(v1), u_tetrad)


    @classmethod
    def vect_type(cls, v1):
        """Return type of bit vector v1 in GF(2)**24. 

        This is either (w,d,y), 0 <= d <= 3,  or (w,4,t)

        Here w is the bit weight of vector v1. In the first case, there
        is a unique syndrome s(v1) of v1 of weight d, 0 <= d <= 3, and
        the weight of v & s(v1) is equal to y in case w <= 12 and to
        d - y in case w >= 12.

        In the second case the syndrome may be any of a tetrad of 
        six vectors s_i of length 4 with disjoint support. In case 
        w = 12, the value  t is the number of vectors s_i completely 
        contained in v1. For other values of w the value t is zero.
  
        It is well known that all vectors of the same type are
        in one orbit of the automorphism group of the Golay code.
        """
        x = v1
        w = cls.bw24(x)
        s = cls.syndrome(x)
        d = cls.bw24(s)
        if d < 4: 
           y =  bw24(x & s) 
           if w >= 12: y = d - y
        elif  w != 12: 
           y = 0  
        else:
           y, rest = 0 + (s == x & s),  x & ~s & 0xffffff 
           while rest:
               s =  cls.syndrome(x, cls.lsbit24(rest))
               y +=  s == x & s  
               rest &= ~s
        return ( w, d, y )


   
    @classmethod
    def gcode_weight(cls, v1):
        """Return bit weight of Golay code word v1 divided by 4

        Here 0 <= v1 < 4096 is the number of a Golay code word.
        """
        t = - ((v1 >> 11) & 1)
        w = (cls.theta_table[v1 & 0x7ff] >> 12) & 7
        return ((w & 7) ^ t) + (t & 7)  

    @classmethod
    def gcode_to_bit_list(cls, v1):
        """Return Golay code word 'v1' as a list of bit positions.

        Here 0 <= v1 < 4096 is the number of a Golay code word.
        """
        v = cls.gcode_to_vect(v1)
        length, bit_list = cls.vect_to_bit_list(v)
        return bit_list[:length]

    @classmethod
    def cocode_weight(cls, c1):
        """Return minimum possible weight of cocode vector c1.
  
        Here c1 is a cocode vector in cocode representation. 
        """
        syn = cls.cocode_syndrome(c1, u_tetrad=0)
        return cls.bw24(syn)

    @classmethod
    def cocode_to_bit_list(cls, c1, u_tetrad=24):
        """Return cocode syndrome of 'c1' as list of bit positions.

        Here c1 is a cocode vector in cocode representation. The 
        function returns the sorted list of the bit positions of 
        the syndrome of c1.  That syndrome is unique if it has 
        length < 4. If the syndrome has length four then we return 
        the following syndrome:
          - A syndrome containing u_tetrad if 0 <= u_tetrad < 24, 
          - We raise ValueError if  'u_tetrad' == 24.  
        """
        v = cls.cocode_syndrome(c1, u_tetrad)
        length, bit_list = cls.vect_to_bit_list(v)
        return bit_list[:length]


    @classmethod
    def cocode_to_sextet(cls, c1):
        """Return a cocode word 'c1' in as a sextet.
 
        Here 'c1' is an cocode word in cocode representation. The 
        function stores concatenation of the 6 tetrads that make up 
        the sextet 'c1' is a list of length 24. The (ordered) 
        tetrads are stored in lexical order. 
        It raises ValueError if 'c1' is not a tetrad.
        """
        sextet = cls.cocode_to_bit_list(c1, 0)
        if len(sextet) != 4:
            raise ValueError("Golay cocode word is not a sextet") 
        v = set(range(24)) - set(sextet)
        while len(v):
            tetrad = cls.cocode_to_bit_list(c1, min(v)) 
            v = v - set(tetrad)
            sextet += tetrad
        return sextet
        


    ############################################################################
    # Scalar product of Golay code and cocode
    ############################################################################


    @classmethod
    def scalar_prod(cls, v1, c1):
        """Return scalar product (v1,c1) of Golay code and cocode vector

        v1 is a Golay code vector in 'gcode' representation, c1 is a
        cocode vector in cocode representation.
        """
        r = v1 & c1
        r ^= r >> 6
        r ^= r >> 3
        return (0x96 >> (r & 7)) & 1


    ############################################################################
    # Conversion from and to suboctads
    ############################################################################

    @classmethod
    def suboctad_to_cocode(cls, u_sub, v1):
        """Convert even suboctad u_sub of octad v to cocode representation

        Each octad v1 has 64 even subsets, when a subset x of v1 is identified
        with its complement v1 \\ x. These subsets are called suboctads.
        Let b_0,...,b_7 be the elements of the octad v1 in natural order. 
        Then the even subset (b_0 , b_i)  has suboctad number 2**(i-1) for 
        i = 1,...,6. Combining suboctads by disjoint union corresponds to 
        combining their numbers by xor. The empty subocatad has number 0.
        This yields a 1:1 correspondence between the integers 0,...,63 and
        the suboctads of a fixed octad v1, when idedentifying a suboctad with
        its complement.

        The function returns the suboctad of v1 with number u_sub in 'cocode'
        representation. Octad v1 must be given in 'gcode' representation.
        An exception is raised if v1 does not represent an octad.
        """
        oct = cls.gcode_to_vect(v1)
        w = cls.bw24(oct)
        if w == 16:
             w, oct =  8, oct ^ 0xffffff
        assert w == 8
        parity =  (0x96 >> ((u_sub ^ (u_sub >> 3)) & 7)) & 1
        sub = parity + ((u_sub & 0x3f) << 1)
        vector =  cls.spread_b24(sub, oct)
        return cls.vect_to_cocode(vector) 
        
    @classmethod
    def cocode_to_suboctad(cls, c1, v1):
        """Convert cocode element c1 to suboctad of octad v1.

        This is an inverse of member function suboctad_to_cocode().
        The function returns the suboctad number corresponding to the
        cocode element c1, if v1 is an octad and c1 is an even subset 
        of v1. c1 and v1 must be given in 'cocode' and 'gcode'
        representation, respectively. The function raises an exception
        if v1 is not an octad or c1 cannot be represented as an even
        subset of v1.
        """ 
        assert c1 & 0x800 == 0
        oct = cls.gcode_to_vect(v1)
        w = cls.bw24(oct)
        if w == 16:
             w, oct =  8, oct ^ 0xffffff
        assert w == 8
        syn = cls.cocode_syndrome(c1, cls.lsbit24(oct))
        assert syn & oct == syn
        u_sub = cls.extract_b24(syn, oct)
        b7 = (u_sub >> 7) & 1
        u_sub ^= (b7 << 8) - b7
        return (u_sub >> 1) & 0x3f


    @staticmethod
    def suboctad_weight(u_sub):
        """Return parity of halved bit weight of suboctad u_sub
    
        See member function suboctad_to_cocode for documentation of 
        suboctads. The function returns 0 is the bit weight of a
        suboctad is divisible by four and 1 othewise.
        """
        w = (u_sub & 0x15) + ((u_sub & 0x2a) >> 1)
        w = w + (w >> 2) + (w >> 4) + 1
        return (w >> 1) & 1 
  
    @staticmethod
    def suboctad_scalar_prod(u_sub1, u_sub2):
        """Return scalar product of suboctads u_sub1, u_sub2
    
        Here the scalar product is the parity of the  vector  
        u_sub1 & u_sub2 when u_sub1 and u_sub2 are given as subsets
        of an octad in vector notation.

        But in this functions parameters u_sub1, u_sub2 are suboctad 
        numbers as described in function suboctad_to_cocode.
        """
        wp = (0x96 >> ((u_sub1 ^ (u_sub1 >> 3)) & 7)) 
        wp &=  (0x96 >> ((u_sub2 ^ (u_sub2 >> 3)) & 7))
        u_sub1 &= u_sub2
        wp ^= (0x96 >> ((u_sub1 ^ (u_sub1 >> 3)) & 7)) 
        return wp & 1





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from random import randint
import types
import sys
import re
import os
from operator import __or__, __xor__
from functools import reduce
import pytest


from mmgroup.bitfunctions import bitlen, bitweight, hibit, v2


        
                        
 
 
from mmgroup.dev.mat24.mat24_ref import Mat24
from mmgroup.dev.mat24.make_addition_table import make_addition_operations
from mmgroup.dev.mat24.make_addition_table import check_addition_operations
from mmgroup.dev.mat24.make_addition_table import display_addition_operations
from mmgroup.dev.mat24.make_addition_table import BitMatrixMulFix


def table_testcases():
    yield Mat24.basis[13:], False, 12
    yield [0], True, 4
    yield [0], False, 4
    for i in range(1, 10):
         mask =  (1 << i) - 1
         for j in range(1, 10):
             m = [randint(0,  mask) for k in range(j)]
             yield m, randint(0,1), randint(4,6)


@pytest.mark.mat24
def test_make_additon_table(verbose =  0):
    print("\nTesting module make_addition_table")
    for i, (m, singleton, granularity)  in enumerate(table_testcases()):
        if verbose:
            nrows, ncols = len(m), bitlen(reduce(__or__, m, 0))
            print("\nTest case %d, a %d time %d matrix, singleton = %s" %
                (i, nrows, ncols, bool(singleton)))
            print("m =", [hex(x) for x in m])
        ops = make_addition_operations(m, singleton, granularity)
        if verbose:
            display_addition_operation(ops) 
        check_addition_operations(m, ops, singleton)

    """
    data = [ 6, 10, 12, 15]
    d, stage = make_addition_tree(data, granularity = 4)  
  
    d, stage = make_addition_tree(b, granularity = 4) 
    if verbose: 
       show_addition_tree(d)


    bm = BitMatrixMulFix(granularity = 4)
    bm.set_fixed_matrix(b)
    bm.selftest()
    bm.set_names("uint32_t", "c", "a")
    if verbose:
        print( bm.make_c_program() )
        print( "===============================================" )
        print( bm.n_registers , "registers needed" )
        print( len(bm.ops), "operations needed" )
    print("Test passed")
    """
    
    
     
    